from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Iterator, List, Literal, Optional, TypedDict, Union, overload

import gql
import graphql.language as l
from latch_sdk_gql.execute import execute
from latch_sdk_gql.utils import _GqlJsonValue, _json_value, _name_node, _parse_selection
from typing_extensions import TypeAlias

from latch.registry.table import Table


class _CatalogExperimentNode(TypedDict):
    id: str
    displayName: str


@dataclass
class _Cache:
    display_name: Optional[str] = None
    tables: Optional[List[Table]] = None


@dataclass(frozen=True)
class Project:
    """Registry project (folder containing :class:`tables <latch.registry.table.Table>`).

    :meth:`~latch.account.Account.list_registry_projects` is the typical way to get a :class:`Project`.
    """

    _cache: _Cache = field(
        default_factory=lambda: _Cache(),
        init=False,
        repr=False,
        hash=False,
        compare=False,
    )

    id: str
    """Unique identifier."""

    def load(self) -> None:
        """(Re-)populate this project instance's cache.

        Future calls to most getters will return immediately without making a network request.

        Always makes a network request.
        """
        data = execute(
            document=gql.gql("""
                query ProjectQuery($id: BigInt!) {
                    catalogProject(id: $id) {
                        id
                        displayName
                        catalogExperimentsByProjectId (
                            condition: {
                                projectId: $id
                                removed: false
                            }
                        ) {
                            nodes {
                                id
                                displayName
                            }
                        }
                    }
                }
                """),
            variables={"id": self.id},
        )["catalogProject"]
        # todo(maximsmol): deal with nonexistent projects

        self._cache.display_name = data["displayName"]

        self._cache.tables = []
        experiments: List[_CatalogExperimentNode] = data[
            "catalogExperimentsByProjectId"
        ]["nodes"]
        for x in experiments:
            cur = Table(x["id"])
            cur._cache.display_name = x["displayName"]
            self._cache.tables.append(cur)

    # get_display_name

    @overload
    def get_display_name(self, *, load_if_missing: Literal[True] = True) -> str: ...

    @overload
    def get_display_name(self, *, load_if_missing: bool) -> Optional[str]: ...

    def get_display_name(self, *, load_if_missing: bool = True) -> Optional[str]:
        """Get the display name of this project.

        This is an opaque string that can contain any valid Unicode data.

        Display names are *not unique* and *must never be used as identifiers*.
        Use :attr:`id` instead.

        Args:
            load_if_missing:
                If true, :meth:`load` the display name if not in cache.
                If false, return `None` if not in cache.

        Returns:
            Display name.
        """
        if self._cache.display_name is None and load_if_missing:
            self.load()

        return self._cache.display_name

    # list_tables

    @overload
    def list_tables(self, *, load_if_missing: Literal[True] = True) -> List[Table]: ...

    @overload
    def list_tables(self, *, load_if_missing: bool) -> Optional[List[Table]]: ...

    def list_tables(self, *, load_if_missing: bool = True) -> Optional[List[Table]]:
        """List Registry tables contained in this project.

        Args:
            load_if_missing:
                If true, :meth:`load` the table list if not in cache.
                If false, return `None` if not in cache.

        Returns:
            Tables in this project.
        """
        if self._cache.tables is None and load_if_missing:
            self.load()

        return self._cache.tables

    @contextmanager
    def update(self, *, reload_on_commit: bool = True) -> Iterator["ProjectUpdate"]:
        """Start an update transaction.

        The transaction will commit when the context manager closes unless an error occurs.

        No changes will occur until the transaction commits.

        The transaction can be cancelled by running :meth:`ProjectUpdate.clear`
        before closing the context manager.

        Args:
            reload_on_commit:
                If true, :meth:`load` this project after the transaction commits.

        Returns:
            Context manager for the new transaction.
        """

        upd = ProjectUpdate(self)
        yield upd
        upd.commit()

        if reload_on_commit:
            self.load()

    def __repr__(self):
        display_name = self.get_display_name(load_if_missing=False)
        if display_name is not None:
            return f"Project(id={self.id}, display_name={display_name})"

        return f"Project(id={self.id})"

    def __str__(self):
        return repr(self)


@dataclass(frozen=True)
class _ProjectTablesUpsertData:
    display_name: str


@dataclass(frozen=True)
class _ProjectTablesDeleteData:
    id: str


_ProjectTablesMutationData: TypeAlias = Union[
    _ProjectTablesUpsertData, _ProjectTablesDeleteData
]


@dataclass(frozen=True)
class ProjectUpdate:
    _table_mutations: List[_ProjectTablesMutationData] = field(
        default_factory=list,
        init=False,
        repr=False,
        hash=False,
        compare=False,
    )

    project: Project

    # upsert table

    def upsert_table(self, display_name: str):
        """Creates a table.

        Not idempotent. Two calls with the same args will create two tables.

        Args:
            display_name: The display name of the new table.
        """
        self._table_mutations.append(_ProjectTablesUpsertData(display_name))

    def _add_table_upserts_selection(
        self, upserts: List[_ProjectTablesUpsertData], mutations: List[l.SelectionNode]
    ):
        if len(upserts) == 0:
            return

        display_names: _GqlJsonValue = [x.display_name for x in upserts]

        res = _parse_selection("""
            catalogMultiCreateExperiments(input: {}) {
                clientMutationId
            }
            """)
        assert isinstance(res, l.FieldNode)

        args = l.ArgumentNode()
        args.name = _name_node("input")
        args.value = _json_value(
            {
                "argProjectId": self.project.id,
                "argDisplayNames": display_names,
            }
        )

        res.alias = _name_node(f"upd{len(mutations)}")
        res.arguments = tuple([args])

        mutations.append(res)

    # delete table

    def delete_table(self, id: str):
        """Deletes a table.

        Args:
            id: The ID of the target table.
        """
        self._table_mutations.append(_ProjectTablesDeleteData(id))

    def _add_table_deletes_selection(
        self, deletes: List[_ProjectTablesDeleteData], mutations: List[l.SelectionNode]
    ):
        if len(deletes) == 0:
            return

        ids: _GqlJsonValue = [x.id for x in deletes]

        res = _parse_selection("""
            catalogMultiDeleteExperiments(input: {}) {
                clientMutationId
            }
            """)
        assert isinstance(res, l.FieldNode)

        args = l.ArgumentNode()
        args.name = _name_node("input")
        args.value = _json_value(
            {
                "argIds": ids,
            }
        )

        res.alias = _name_node(f"upd{len(mutations)}")
        res.arguments = tuple([args])

        mutations.append(res)

    # transaction

    def commit(self) -> None:
        """Commit this project update transaction.

        May be called multiple times.

        All pending updates are committed with one network request.

        Atomic. The entire transaction either commits or fails with an exception.
        """
        mutations: List[l.SelectionNode] = []

        if len(self._table_mutations) == 0:
            return

        def _add_table_mutations_selection(cur):
            if isinstance(cur[0], _ProjectTablesUpsertData):
                self._add_table_upserts_selection(cur, mutations)
            if isinstance(cur[0], _ProjectTablesDeleteData):
                self._add_table_deletes_selection(cur, mutations)

        cur = [self._table_mutations[0]]
        for mut in self._table_mutations[1:]:
            if isinstance(mut, type(cur[0])):
                cur.append(mut)
                continue

            _add_table_mutations_selection(cur)
            cur = [mut]

        _add_table_mutations_selection(cur)

        sel_set = l.SelectionSetNode()
        sel_set.selections = tuple(mutations)

        doc = l.parse("""
            mutation ProjectUpdate {
                placeholder
            }
        """)

        assert len(doc.definitions) == 1
        mut = doc.definitions[0]

        assert isinstance(mut, l.OperationDefinitionNode)
        mut.selection_set = sel_set

        execute(doc)

        self.clear()

    def clear(self):
        """Remove pending updates.

        May be called to cancel any pending updates that have not been committed.
        """
        self._table_mutations.clear()



import subprocess
from pathlib import Path

from latch.functions.messages import message
from latch.resources.tasks import small_task
from latch.types.directory import LatchOutputDir
from latch.types.file import LatchFile


@small_task
def assembly_task(
    read1: LatchFile, read2: LatchFile, output_directory: LatchOutputDir
) -> LatchFile:
    # A reference to our output.
    sam_file = Path("covid_assembly.sam").resolve()

    bowtie2_cmd = [
        "bowtie2/bowtie2",
        "--local",
        "--very-sensitive-local",
        "-x",
        "wuhan",
        "-1",
        read1.local_path,
        "-2",
        read2.local_path,
        "-S",
        str(sam_file),
    ]

    try:
        # We use shell=True for all the benefits of pipes and other shell features.
        # When using shell=True, we pass the entire command as a single string as
        # opposed to a list since the shell will parse the string into a list
        # using its own rules.
        subprocess.run(" ".join(bowtie2_cmd), shell=True, check=True)
    except subprocess.CalledProcessError as e:
        # will display in the messages tab of the execution graph for the assembly_task node
        message(
            "error",
            {"title": "Bowtie2 Failed", "body": f"Error: {str(e)}"},
        )
        raise e

    # intended output path of the file in Latch console, constructed from
    # the user provided output directory
    output_location = f"{output_directory.remote_directory}/covid_assembly.sam"

    return LatchFile(str(sam_file), output_location)



import subprocess
from typing import List

from latch.resources.tasks import small_task
from latch.types.directory import LatchOutputDir
from latch.types.file import LatchFile


def run(command: List[str], check: bool = True, capture_output: bool = False):
    cmd = ["conda", "run", "--name", "example", "/bin/bash", "-c", " ".join(command)]

    return subprocess.run(cmd, check=check, capture_output=capture_output)


# change the name of this function to something more descriptive
@small_task
def conda_task(input_file: LatchFile, output_directory: LatchOutputDir) -> LatchFile:
    # You can run conda packages as a subprocess:

    run(
        [
            "put",
            "conda",
            "command",
            "here",
        ]
    )
    ...
    output_location = f"{output_directory.remote_directory}/YOUR_OUTPUT_FILE_NAME"
    return LatchFile(str("/root/LOCAL_FILE_IN_TASK"), output_location)



import json
import os
import platform
import sys
import tarfile
import tempfile
from dataclasses import asdict, dataclass
from pathlib import Path
from traceback import print_exc
from types import TracebackType
from typing import List, Optional, Type

import click

from latch_cli.constants import latch_constants
from latch_cli.exceptions.traceback import _Traceback
from latch_cli.utils import get_local_package_version


@dataclass(frozen=True)
class _Metadata:
    os_name: str = os.name
    os_version: str = platform.version()
    latch_version: str = get_local_package_version()
    py_version: str = sys.version
    platform: str = platform.platform()

    def print(self):
        click.secho("Crash info:", fg="red", bold=True)
        click.echo(
            " ".join(
                [
                    click.style("Latch SDK version:", fg="red"),
                    self.latch_version,
                ]
            )
        )
        click.echo(
            " ".join(
                [
                    click.style("Python version:", fg="red"),
                    self.py_version.replace("\n", ";"),
                ]
            )
        )
        click.echo(" ".join([click.style("Platform:", fg="red"), self.platform]))
        click.echo(
            " ".join(
                [
                    click.style("OS:", fg="red"),
                    f"{self.os_name}; {self.os_version}",
                ]
            )
        )


class CrashHandler:
    """Display and store useful information when the program fails

    * Display tracebacks
    * Parse and display opaque flytekit serialization error messages
    * Write necessary information to reproduce failure to a tarball
    """

    def __init__(self):
        self.metadata: _Metadata = _Metadata()
        self.message: Optional[str] = None
        self.pkg_root: Optional[str] = None

    def _write_state_to_tarball(self):
        """Bundles files needed to reproduce failed state into a tarball.

        Tarball contains:
          * JSON holding platform + package version metadata
          * logs directory holding docker build logs
          * Text file holding traceback
          * (If register) workflow package (python files, Dockerfile, etc.)
        """

        tarball_path = Path(".latch_report.tar.gz").resolve()
        tarball_path.unlink(missing_ok=True)

        with tarfile.open(tarball_path, mode="x:gz") as tf:
            # If calling stack frame is handling an exception, we want to store
            # the traceback in a log file.
            if sys.exc_info()[0] is not None:
                with tempfile.NamedTemporaryFile("w+") as ntf:
                    print_exc(file=ntf)
                    ntf.seek(0)
                    tf.add(ntf.name, arcname="traceback.txt")

            if self.pkg_root is not None:
                logs_path = Path(self.pkg_root) / ".latch" / ".logs"
                if logs_path.exists():
                    tf.add(logs_path, arcname="logs")

                pkg_files: List[Path] = []
                for dp, _, fnames in os.walk(logs_path):
                    for f in fnames:
                        p = (Path(dp) / f).resolve()
                        if (
                            latch_constants.ignore_regex.search(str(p))
                            or p.is_symlink()
                            or p.stat().st_size > latch_constants.file_max_size
                        ):
                            continue

                        pkg_files.append(p)

                for file_path in pkg_files:
                    tf.add(file_path)

            with tempfile.NamedTemporaryFile("w+") as ntf:
                json.dump(asdict(self.metadata), ntf)
                ntf.seek(0)
                tf.add(ntf.name, arcname="metadata.json")

        print(f"Crash report written to {tarball_path}")

    def init(self):
        """Custom error handling.

        When an exception is thrown:
            - Display an optional message
            - Pretty print the traceback
            - Parse unintuitive errors and redisplay
            - Store useful system information in a tarball for debugging
        """

        def _excepthook(
            type_: Type[BaseException],
            value: BaseException,
            traceback: Optional[TracebackType],
        ) -> None:
            click.secho(f"\n{self.message}:\n", fg="red", bold=True)
            _Traceback(type_, value, traceback).pretty_print()

            self.metadata.print()

            if os.environ.get("LATCH_NO_CRASH_REPORT") == "1":
                click.secho(
                    "Not generating crash report due to $LATCH_NO_CRASH_REPORT",
                    bold=True,
                )
                return

            if not click.confirm("Generate a crash report?"):
                return

            print("Generating...")
            self._write_state_to_tarball()

        sys.excepthook = _excepthook



from flytekit.core.condition import ConditionalSection, conditional


def create_conditional_section(name: str) -> ConditionalSection:
    """
    This method creates a new conditional section in a workflow, allowing a user
    to conditionally execute a task based on the value of a task result.

    Conditional sections are akin to ternary operators -- they return the output
    of the branch result. However, they can be n-ary with as many _elif_ clauses
    as desired.

    It is possible to consume the outputs from conditional nodes. And to pass in
    outputs from other tasks to conditional nodes.

    The boolean expressions in the condition use `&` and `|` as and / or operators.
    Additionally, unary expressions are not allowed. Thus if a task returns a boolean
    and we wish to use it in a condition of a conditional block, we must
    use built in truth checks: `result.is_true()` or `result.is_false()`

    Args:
        name: The name of the conditional section, to be shown in Latch Console

    Returns:
        A conditional section

    Intended Use: ::

        @workflow
        def multiplier(my_input: float) -> float:
            result_1 = double(n=my_input)
            result_2 =  (
                create_conditional_section("fractions")
                .if_((result_1 < 0.0)).then(double(n=result_1))
                .elif_((result_1 > 0.0)).then(square(n=result_1))
                .else_().fail("Only nonzero values allowed")
            )
            result_3 = double(n=result_2)
            return result_3
    """
    return conditional(name)



"""Authorization utilities."""

import base64


def _base64url_encode(bytestring: bytes) -> str:
    """Construct a "base64url" encoding of a bytestring.

    Not that this is a modified Base64 encoding with a url + filename safe
    alphabet. This "base64url" encoding is *not* the same thing as a
    "base64" encoding.

    The differences are enumerated in the `OAuth2 RFC`_.

    Args:
        bytestring: string of bytes to be encoded

    Returns:
        Encoded bytestring

    .. _OAuth2 RFC:
        https://datatracker.ietf.org/doc/html/rfc4648#section-5
    """
    # Padding must be stripped as specified in RFC.
    return base64.urlsafe_b64encode(bytestring).decode("utf-8").replace("=", "")



from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import date, datetime
from enum import Enum
from typing import (
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    TypedDict,
    Union,
    cast,
    get_args,
    get_origin,
    overload,
)

import gql
import graphql.language as l
import graphql.language.parser as lp
from latch_sdk_gql.execute import execute
from latch_sdk_gql.utils import (
    _GqlJsonValue,
    _json_value,
    _name_node,
    _parse_selection,
    _var_def_node,
    _var_node,
)
from typing_extensions import TypeAlias

from latch.registry.record import NoSuchColumnError, Record
from latch.registry.types import (
    Column,
    InvalidValue,
    LinkedRecordType,
    RecordValue,
    RegistryEnumDefinition,
    RegistryPythonType,
    RegistryPythonValue,
)
from latch.registry.upstream_types.types import DBType, RegistryType
from latch.registry.upstream_types.values import DBValue, EmptyCell
from latch.registry.utils import to_python_literal, to_python_type, to_registry_literal
from latch.types.directory import LatchDir
from latch.types.file import LatchFile

from ..types.json import JsonValue


class _AllRecordsNode(TypedDict):
    sampleId: str
    sampleName: str
    sampleDataKey: str
    sampleDataValue: DBValue


class _ColumnNode(TypedDict("_ColumnNodeReserved", {"def": DBValue})):
    key: str
    type: DBType


@dataclass
class _Cache:
    display_name: Optional[str] = None
    columns: Optional[Dict[str, Column]] = None
    project_id: Optional[str] = None


@dataclass(frozen=True)
class Table:
    """Registry table. Contains :class:`records <latch.registry.record.Record>`.

    :meth:`~latch.registry.project.Project.list_tables` is the typical way to get a :class:`Table`.
    """

    _cache: _Cache = field(
        default_factory=lambda: _Cache(),
        init=False,
        repr=False,
        hash=False,
        compare=False,
    )

    id: str
    """Unique identifier."""

    def load(self) -> None:
        """(Re-)populate this table instance's cache.

        Future calls to most getters will return immediately without making a network request.

        Always makes a network request.
        """
        data = execute(
            gql.gql("""
                query TableQuery($id: BigInt!) {
                    catalogExperiment(id: $id) {
                        id
                        displayName
                        catalogExperimentColumnDefinitionsByExperimentId {
                            nodes {
                                key
                                type
                                def
                            }
                        }
                        projectId
                    }
                }
                """),
            variables={"id": self.id},
        )["catalogExperiment"]
        # todo(maximsmol): deal with nonexistent tables

        self._cache.project_id = data["projectId"]
        self._cache.display_name = data["displayName"]

        self._cache.columns = {}
        columns: List[_ColumnNode] = data[
            "catalogExperimentColumnDefinitionsByExperimentId"
        ]["nodes"]
        for x in columns:
            py_type = to_python_type(x["type"]["type"])
            if x["type"]["allowEmpty"]:
                py_type = Union[py_type, EmptyCell]

            cur = Column(x["key"], py_type, x["type"])
            self._cache.columns[cur.key] = cur

    # get_project_id

    @overload
    def get_project_id(self, *, load_if_missing: Literal[True] = True) -> str: ...

    @overload
    def get_project_id(self, *, load_if_missing: bool) -> Optional[str]: ...

    def get_project_id(self, *, load_if_missing: bool = True) -> Optional[str]:
        """Get the ID of the project that contains this table.

        Args:
            load_if_missing:
                If true, :meth:`load` the project ID if not in cache.
                If false, return `None` if not in cache.

        Returns:
            ID of the :class:`Project` containing this table.
        """
        if self._cache.project_id is None:
            if not load_if_missing:
                return None

            self.load()

        return self._cache.project_id

    # get_display_name

    @overload
    def get_display_name(self, *, load_if_missing: Literal[True] = True) -> str: ...

    @overload
    def get_display_name(self, *, load_if_missing: bool) -> Optional[str]: ...

    def get_display_name(self, *, load_if_missing: bool = True) -> Optional[str]:
        """Get the display name of this table.

        This is an opaque string that can contain any valid Unicode data.

        Display names are *not unique* and *must never be used as identifiers*.
        Use :attr:`id` instead.

        Args:
            load_if_missing:
                If true, :meth:`load` the display name if not in cache.
                If false, return `None` if not in cache.

        Returns:
            Display name.
        """
        if self._cache.display_name is None and load_if_missing:
            self.load()

        return self._cache.display_name

    # get_columns

    @overload
    def get_columns(
        self, *, load_if_missing: Literal[True] = True
    ) -> Dict[str, Column]: ...

    @overload
    def get_columns(self, *, load_if_missing: bool) -> Optional[Dict[str, Column]]: ...

    def get_columns(
        self, *, load_if_missing: bool = True
    ) -> Optional[Dict[str, Column]]:
        """Get the columns of this table.

        Args:
            load_if_missing:
                If true, :meth:`load` the column list if not in cache.
                If false, return `None` if not in cache.

        Returns:
            Mapping between column keys and :class:`columns <latch.registry.types.Column>`.
        """
        if self._cache.columns is None and load_if_missing:
            self.load()

        return self._cache.columns

    def list_records(self, *, page_size: int = 100) -> Iterator[Dict[str, Record]]:
        """List Registry records contained in this table.

        Args:
            page_size:
                Maximum size of a page of records. The last page may be shorter
                than this value.

        Yields:
            Pages of records. Each page is a mapping between record IDs and
            :class:`records <latch.registry.record.Record>`.
        """

        cols = self.get_columns()

        # todo(maximsmol): because allSamples returns each column as its own
        # row, we can't paginate by samples because we don't know when a sample is finished
        nodes: List[_AllRecordsNode] = execute(
            gql.gql("""
                query TableQuery($id: BigInt!) {
                    catalogExperiment(id: $id) {
                        allSamples {
                            nodes {
                                sampleId
                                sampleName
                                sampleDataKey
                                sampleDataValue
                            }
                        }
                    }
                }
                """),
            {
                "id": self.id,
            },
        )["catalogExperiment"]["allSamples"]["nodes"]
        # todo(maximsmol): deal with nonexistent tables

        record_names: Dict[str, str] = {}
        record_values: Dict[str, Dict[str, RecordValue]] = {}

        for node in nodes:
            record_names[node["sampleId"]] = node["sampleName"]
            vals = record_values.setdefault(node["sampleId"], {})

            col = cols.get(node["sampleDataKey"])
            if col is None:
                continue

            # todo(maximsmol): in the future, allow storing or yielding values that failed to parse
            vals[col.key] = to_python_literal(
                node["sampleDataValue"], col.upstream_type["type"]
            )

        page: Dict[str, Record] = {}
        for id, values in record_values.items():
            for col in cols.values():
                if col.key in values:
                    continue

                if not col.upstream_type["allowEmpty"]:
                    values[col.key] = InvalidValue("")

            cur = Record(id)
            cur._cache.name = record_names[id]
            cur._cache.values = values
            cur._cache.columns = cols
            page[id] = cur

            if len(page) == page_size:
                yield page
                page = {}

        if len(page) > 0:
            yield page

    def get_dataframe(self):
        """Get a pandas DataFrame of all records in this table.

        Returns:
            DataFrame representing all records in this table.
        """

        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                "pandas needs to be installed to use get_dataframe. Install it with"
                " `pip install pandas` or `pip install latch[pandas]`."
            )

        records = []
        for page in self.list_records():
            for record in page.values():
                full_record = record.get_values()
                if full_record is not None:
                    full_record["Name"] = record.get_name()
                    records.append(full_record)

        if len(records) == 0:
            cols = self.get_columns()
            if cols is None:
                return pd.DataFrame()

            return pd.DataFrame(columns=list(cols.keys()))

        return pd.DataFrame(records)

    @contextmanager
    def update(self, *, reload_on_commit: bool = True) -> Iterator["TableUpdate"]:
        """Start an update transaction.

        The transaction will commit when the context manager closes unless an error occurs.

        No changes will occur until the transaction commits.

        The transaction can be cancelled by running :meth:`TableUpdate.clear`
        before closing the context manager.

        Args:
            reload_on_commit:
                If true, :meth:`load` this table after the transaction commits.

        Returns:
            Context manager for the new transaction.
        """

        upd = TableUpdate(self)
        yield upd
        upd.commit()

        if reload_on_commit:
            self.load()

    def __repr__(self):
        display_name = self.get_display_name(load_if_missing=False)
        if display_name is not None:
            return f"Table(id={self.id}, display_name={display_name})"

        return f"Table(id={self.id})"

    def __str__(self):
        return repr(self)


@dataclass(frozen=True)
class _TableRecordsUpsertData:
    name: str
    values: Dict[str, DBValue]


@dataclass(frozen=True)
class _TableRecordsDeleteData:
    name: str


@dataclass(frozen=True)
class _TableColumnUpsertData:
    key: str
    type: DBType


_TableRecordsMutationData: TypeAlias = Union[
    _TableRecordsUpsertData,
    _TableRecordsDeleteData,
    _TableColumnUpsertData,
]


class InvalidColumnTypeError(ValueError):
    """Failure to use an invalid column type.

    Attributes:
        key: Identifier of the invalid column.
        invalid_type: Requested column type.
    """

    def __init__(
        self, key: str, invalid_type: Union[Type[object], RegistryPythonType], msg: str
    ):
        super().__init__(
            f"invalid column type for {repr(key)}. {msg}: {repr(invalid_type)}"
        )

        self.key = key
        self.invalid_type = invalid_type


@dataclass(frozen=True)
class TableUpdate:
    """Ongoing :class:`Table` update transaction.

    Groups requested updates to commit everything together in one network request.

    Transactions are atomic. The entire transaction either commits or fails with an exception.
    """

    _record_mutations: List[_TableRecordsMutationData] = field(
        default_factory=list,
        init=False,
        repr=False,
        hash=False,
        compare=False,
    )

    table: Table

    # upsert record

    def upsert_record_raw_unsafe(
        self, *, name: str, values: Dict[str, DBValue]
    ) -> None:
        """DANGEROUSLY Update or create a record using raw :class:`values <latch.registry.upstream_types.values.DBValue>`.

        Values are not checked against the existing columns.

        A transport error will be thrown if non-existent columns are updated.

        The update will succeed if values do not match column types and future
        reads will produce :class:`invalid values <latch.registry.types.InvalidValue>`.

        Unsafe:
            The value dictionary is not validated in any way.
            It is possible to create completely invalid record values that
            are not a valid Registry value of any type. Future reads will
            fail catastrophically when trying to parse these values.

        Args:
            name: Target record name.
            values: Column values that to set.
        """
        self._record_mutations.append(_TableRecordsUpsertData(name, values))

    def upsert_record(self, name: str, **values: RegistryPythonValue) -> None:
        """Update or create a record.

        A transport error will be thrown if non-existent columns are updated.

        It is possible that the column definitions changed since the table was last
        loaded and the update will be issued with values that do not match current column types.
        This will succeed with no error and future reads will produce :class:`invalid values <latch.registry.types.InvalidValue>`.

        Args:
            name: Target record name.
            values: Column values to set.
        """
        cols = self.table.get_columns()

        db_vals: Dict[str, DBValue] = {}
        for k, v in values.items():
            col = cols.get(k)
            if col is None:
                raise NoSuchColumnError(k)

            db_vals[k] = to_registry_literal(v, col.upstream_type["type"])

        self._record_mutations.append(_TableRecordsUpsertData(name, db_vals))

    def _add_record_upserts_selection(
        self,
        upserts: List[_TableRecordsUpsertData],
        mutations: List[l.SelectionNode],
        vars: Dict[str, Tuple[l.TypeNode, JsonValue]],
    ) -> None:
        if len(upserts) == 0:
            return

        names: _GqlJsonValue = [x.name for x in upserts]
        values: JsonValue = [cast(Dict[str, JsonValue], x.values) for x in upserts]

        res = _parse_selection("""
            catalogMultiUpsertSamples(input: {}) {
                clientMutationId
            }
        """)
        assert isinstance(res, l.FieldNode)

        argDataVar = f"upd{len(mutations)}ArgData"

        args = l.ArgumentNode()
        args.name = _name_node("input")
        args.value = _json_value(
            {
                "argExperimentId": self.table.id,
                "argNames": names,
                "argData": _var_node(argDataVar),
            }
        )

        res.alias = _name_node(f"upd{len(mutations)}")
        res.arguments = tuple([args])

        mutations.append(res)
        vars[argDataVar] = (l.parse_type("[JSON]"), values)

    # delete record

    def delete_record(self, name: str) -> None:
        """Delete a record.

        Args:
            name: Target record name.
        """
        self._record_mutations.append(_TableRecordsDeleteData(name))

    def _add_record_deletes_selection(
        self, deletes: List[_TableRecordsDeleteData], mutations: List[l.SelectionNode]
    ) -> None:
        if len(deletes) == 0:
            return

        names: _GqlJsonValue = [x.name for x in deletes]

        res = _parse_selection("""
            catalogMultiDeleteSampleByName(input: {}) {
                clientMutationId
            }
            """)
        assert isinstance(res, l.FieldNode)

        args = l.ArgumentNode()
        args.name = _name_node("input")
        args.value = _json_value(
            {
                "argExperimentId": self.table.id,
                "argNames": names,
            }
        )

        res.alias = _name_node(f"upd{len(mutations)}")
        res.arguments = tuple([args])

        mutations.append(res)

    # upsert column

    def upsert_column(
        self,
        key: str,
        type: RegistryPythonType,
        *,
        required: bool = False,
    ):
        """Create a column. Support for updating columns is planned.

        Args:
            key: Identifier of the new column.
            type:
                Type of the new column.

                Only a limited set of Python types is currently supported and
                will be expanded over time.

                :class:`latch.registry.types.RegistryPythonType` represents the currently supported types.
            required:
                If true, records without a value for this column are considered invalid.

                Note that an explicit `None` value is different from a missing/empty value.
                `None` is a valid value for an `Optional` (nullable) column marked as required.

        """

        registry_type: Optional[RegistryType] = None
        if type is str:
            registry_type = {"primitive": "string"}
        if type is int:
            registry_type = {"primitive": "integer"}
        if type is float:
            registry_type = {"primitive": "number"}
        if type is date:
            registry_type = {"primitive": "date"}
        if type is datetime:
            registry_type = {"primitive": "datetime"}
        if type is bool:
            registry_type = {"primitive": "boolean"}
        if type is LatchFile:
            registry_type = {"primitive": "blob"}
        if type is LatchDir:
            registry_type = {"primitive": "blob", "metadata": {"nodeType": "dir"}}

        origin = get_origin(type)
        if origin is not None:
            if issubclass(origin, List):
                inner_type = get_args(type)[0]
                inner_type_origin = get_origin(inner_type)

                if inner_type_origin is not None:
                    if issubclass(inner_type_origin, LinkedRecordType):
                        experiment_id = get_args(get_args(inner_type)[0])[0]
                        registry_type = {
                            "array": {
                                "primitive": "link",
                                "experimentId": experiment_id,
                            }
                        }
                    else:
                        raise InvalidColumnTypeError(
                            key, type, "Unsupported list inner type"
                        )
                elif issubclass(inner_type, LatchFile):
                    registry_type = {"array": {"primitive": "blob"}}
                elif issubclass(inner_type, LatchDir):
                    registry_type = {
                        "array": {"primitive": "blob", "metadata": {"nodeType": "dir"}}
                    }
                else:
                    raise InvalidColumnTypeError(
                        key, type, "Unsupported list inner type"
                    )

            if issubclass(origin, LinkedRecordType):
                experiment_id = get_args(get_args(type)[0])[0]
                registry_type = {"primitive": "link", "experimentId": experiment_id}

            if issubclass(origin, RegistryEnumDefinition):
                members = list(get_args(t)[0] for t in get_args(get_args(type)[0]))
                for x in members:
                    if isinstance(x, str):
                        continue
                    raise InvalidColumnTypeError(
                        key, type, f"Enum value {repr(x)} is not a string"
                    )

                registry_type = {
                    "primitive": "enum",
                    "members": members,
                }

        if isinstance(type, Enum):
            members: List[str] = []
            for f in cast(Type[Enum], type):
                if not isinstance(f.value, str):
                    raise InvalidColumnTypeError(
                        key,
                        type,
                        f"Enum value for {repr(f.name)} ({repr(f.value)}) is not a"
                        " string",
                    )

                members.append(f.value)

            registry_type = {
                "primitive": "enum",
                "members": members,
            }

        if registry_type is None:
            raise InvalidColumnTypeError(key, type, "Unsupported type")

        db_type: DBType = {"type": registry_type, "allowEmpty": not required}
        self._record_mutations.append(_TableColumnUpsertData(key, db_type))

    def _add_column_upserts_selection(
        self,
        upserts: List[_TableColumnUpsertData],
        mutations: List[l.SelectionNode],
        vars: Dict[str, Tuple[l.TypeNode, JsonValue]],
    ) -> None:
        if len(upserts) == 0:
            return

        keys: _GqlJsonValue = [x.key for x in upserts]
        types: JsonValue = [cast(JsonValue, x.type) for x in upserts]

        res = _parse_selection("""
            catalogExperimentColumnDefinitionMultiUpsert(input: {}) {
                clientMutationId
            }
        """)
        assert isinstance(res, l.FieldNode)

        argTypesVar = f"upd{len(mutations)}ArgTypes"

        args = l.ArgumentNode()
        args.name = _name_node("input")
        args.value = _json_value(
            {
                "argExperimentId": self.table.id,
                "argKeys": keys,
                "argTypes": _var_node(argTypesVar),
            }
        )

        res.alias = _name_node(f"upd{len(mutations)}")
        res.arguments = tuple([args])

        mutations.append(res)
        vars[argTypesVar] = (l.parse_type("[JSON]!"), types)

    # transaction

    def commit(self) -> None:
        """Commit this table update transaction.

        May be called multiple times.

        All pending updates are committed with one network request.

        Atomic. The entire transaction either commits or fails with an exception.
        """
        mutations: List[l.SelectionNode] = []
        vars: Dict[str, Tuple[l.TypeNode, JsonValue]] = {}

        if len(self._record_mutations) == 0:
            return

        def _add_record_data_selection(cur):
            if isinstance(cur[0], _TableRecordsUpsertData):
                self._add_record_upserts_selection(cur, mutations, vars)
            if isinstance(cur[0], _TableRecordsDeleteData):
                self._add_record_deletes_selection(cur, mutations)
            if isinstance(cur[0], _TableColumnUpsertData):
                self._add_column_upserts_selection(cur, mutations, vars)

        cur = [self._record_mutations[0]]
        for mut in self._record_mutations[1:]:
            if isinstance(mut, type(cur[0])):
                cur.append(mut)
                continue

            _add_record_data_selection(cur)
            cur = [mut]

        _add_record_data_selection(cur)

        sel_set = l.SelectionSetNode()
        sel_set.selections = tuple(mutations)

        doc = l.parse("""
            mutation TableUpdate {
                placeholder
            }
        """)

        assert len(doc.definitions) == 1
        mut = doc.definitions[0]

        assert isinstance(mut, l.OperationDefinitionNode)
        mut.selection_set = sel_set

        mut.variable_definitions = tuple(
            _var_def_node(k, t) for k, (t, _) in vars.items()
        )

        # todo(maximsmol): catch errors here and raise appropriate Python exceptions
        # 1. column upsert: already exists
        execute(doc, {k: v for k, (_, v) in vars.items()})

        self.clear()

    def clear(self):
        """Remove pending updates.

        May be called to cancel any pending updates that have not been committed.
        """
        self._record_mutations.clear()



import http.server
import json
import ssl
import urllib
import urllib.request
import webbrowser
from typing import Optional

import certifi

from latch_cli.auth.csrf import CSRFState
from latch_cli.auth.pkce import PKCE
from latch_cli.constants import OAuth2Constants


class OAuth2:
    """An object to facilitate the OAuth2.0 flow.

    This implementation of OAuth2.0 follows `RFC5849`_ as gospel.

    (Note we are using the PKCE extension of OAuth2, but the outline of the
    flow below is still a correct model if crytography is removed.)

    ::

         +--------+                               +---------------+
         |        |--(A)- Authorization Request ->|   Resource    |
         |        |                               |     Owner     |
         |        |<-(B)-- Authorization Grant ---|               |
         |        |                               +---------------+
         |        |
         |        |                               +---------------+
         |        |--(C)-- Authorization Grant -->| Authorization |
         | Client |                               |     Server    |
         |        |<-(D)----- Access Token -------|               |
         |        |                               +---------------+
         |        |
         |        |                               +---------------+
         |        |--(E)----- Access Token ------>|    Resource   |
         |        |                               |     Server    |
         |        |<-(F)--- Protected Resource ---|               |
         +--------+                               +---------------+

                         Figure 1: Abstract Protocol Flow



    Note the correspondence between diagram letters and comments in the code
    example snippet below ::

        # Note these context managers hold values critical to flows.
        with PKCE() as pkce:
            with CSRFState() as csrf_state:

                # Construct our object + call each leg of the flow as a method.
                oauth2_flow = OAuth2(pkce, csrf_state, oauth2_constants)
                auth_code = oauth2_flow.authorization_request() # A + B
                token = oauth2_flow.access_token_request(auth_code) # C + D

                # With token, we can do E + F...


    Args:
        pkce: Object managing PKCE values.
        csrf_state: Object managing state for CSRF mitigation.
        oauth2_constants: Object holding constants to identify Latch's authz server.

    .. _RFC5849:
        https://datatracker.ietf.org/doc/html/rfc6749
    """

    def __init__(
        self, pkce: PKCE, csrf_state: CSRFState, oauth2_constants: OAuth2Constants
    ):
        self.pkce = pkce
        self.csrf_state = csrf_state
        self.client_id = oauth2_constants.client_id
        self.authz_server_host = oauth2_constants.authz_server_host
        self.redirect_url = oauth2_constants.redirect_url

    def authorization_request(self, connection: Optional[str]) -> str:
        """Request authorization code from Latch authz server.

        Returns:
            An authorization code to complete the first leg of 0Auth2.0.
        """

        class _CallbackHandler(http.server.BaseHTTPRequestHandler):
            def do_GET(self):
                parsed_url = urllib.parse.urlsplit(self.path)
                if parsed_url.path != "/callback":
                    return

                parsed_query = urllib.parse.parse_qs(parsed_url.query)
                try:
                    code, state = parsed_query["code"][0], parsed_query["state"][0]
                except (KeyError, IndexError):
                    return

                self.send_response(200)
                self.send_header("Content-type", "html")
                self.send_header("Content-Disposition", "inline")
                self.end_headers()

                if state != self.server.csrf_state:
                    # TODO: log csrf attack for debugging but do not fail
                    return

                self.server.authorized = True
                self.server.code = code
                self.server.state = state

                auth_response_html = """
                <!DOCTYPE html>
                <html lang="en">
                  <head>
                    <title>Success</title>
                  </h/ead>
                  <body>
                    <h1>Success</h1>
                    <p>Your Latch SDK has been authenticated.</p>
                  </body>
                </html>
                """.encode("utf-8")
                self.wfile.write(auth_response_html)

            def log_request(self, format, *args):
                """Quiets server."""
                ...

        url_parameters = {
            "scope": "openid profile email",
            "response_type": "code",
            "redirect_uri": self.redirect_url,
            "client_id": self.client_id,
            "code_challenge": self.pkce.challenge,
            "code_challenge_method": self.pkce.challenge_method,
            "state": self.csrf_state.state,
        }
        if connection is not None:
            url_parameters["connection"] = connection

        base_url = self.authz_server_host + "/authorize?"
        url = base_url + urllib.parse.urlencode(url_parameters)
        webbrowser.open_new(url)

        server_name = ("", 5050)
        server = http.server.HTTPServer(server_name, _CallbackHandler)
        server.authorized = False
        server.csrf_state = self.csrf_state.state
        while not server.authorized:
            server.handle_request()

        return server.code

    def access_token_request(self, auth_code: str) -> str:
        """Using a valid code returned from our authz server, request token.

        Args:
            auth_code: Returned from our authz server if it likes us.
        Returns:
            An access token that a user can use to access their resources on
                latch (register workflows, upload files, etc.)
        """

        token_url = self.authz_server_host + "/oauth/token"
        token_body: bytes = json.dumps(
            {
                "grant_type": "authorization_code",
                "client_id": self.client_id,
                "code_verifier": self.pkce.verifier,
                "code": auth_code,
                "redirect_uri": self.redirect_url,
            }
        ).encode("utf-8")
        token_request = urllib.request.Request(token_url, token_body)
        token_request.add_header("Content-Type", "application/json")
        with urllib.request.urlopen(
            token_request, context=ssl.create_default_context(cafile=certifi.where())
        ) as f:
            try:
                json_response = json.loads(f.read().decode("utf-8"))
                id_token = json_response["id_token"]
            except KeyError:
                raise ValueError(f"Passed response is not json {json_response}")

        return id_token



import math
import mimetypes
import os
import random
import time
from concurrent.futures import Future, ProcessPoolExecutor, as_completed, wait
from contextlib import closing
from dataclasses import dataclass
from http.client import HTTPException
from multiprocessing.managers import DictProxy, ListProxy
from pathlib import Path
from queue import Queue
from typing import TYPE_CHECKING, List, Optional, TypedDict

from latch_sdk_config.latch import config as latch_config
from typing_extensions import TypeAlias

from latch.ldata.type import LatchPathError, LDataNodeType
from latch_cli import tinyrequests
from latch_cli.constants import latch_constants, units
from latch_cli.utils import get_auth_header, urljoins, with_si_suffix
from latch_cli.utils.path import normalize_path

from .manager import TransferStateManager
from .node import get_node_data
from .progress import Progress, ProgressBars
from .throttle import Throttle
from .utils import get_max_workers

if TYPE_CHECKING:
    PathQueueType: TypeAlias = "Queue[Optional[Path]]"
    LatencyQueueType: TypeAlias = "Queue[Optional[float]]"
    PartsBySrcType: TypeAlias = DictProxy[Path, ListProxy["CompletedPart"]]
    UploadInfoBySrcType: TypeAlias = DictProxy[Path, "StartUploadReturnType"]


class StartUploadData(TypedDict):
    upload_id: str
    urls: List[str]


@dataclass(frozen=True)
class UploadJob:
    src: Path
    dest: str


@dataclass(frozen=True)
class UploadResult:
    num_files: int
    total_bytes: int
    total_time: int


def upload(
    src: str,  # pathlib.Path strips trailing slashes but we want to keep them here as they determine cp behavior
    dest: str,
    progress: Progress,
    verbose: bool,
    create_parents: bool = False,
) -> UploadResult:
    src_path = Path(src)
    if not src_path.exists():
        raise ValueError(f"could not find {src_path}: no such file or directory.")

    normalized = normalize_path(dest)

    node_data = get_node_data(dest, allow_resolve_to_parent=True)
    assert dest in node_data.data
    dest_data = node_data.data[dest]

    if not (dest_data.exists() or dest_data.is_direct_parent()) and not create_parents:
        raise LatchPathError("no such Latch file or directory", dest)

    dest_is_dir = dest_data.type in {
        LDataNodeType.account_root,
        LDataNodeType.mount,
        LDataNodeType.dir,
    }

    if not dest_is_dir:
        if not dest_data.exists():  # path is latch:///a/b/file_1/file_2
            raise ValueError(f"no such file or directory: {dest}")
        if src_path.is_dir():
            raise ValueError(f"{normalized} is not a directory.")

    if dest.endswith("/") and not dest_data.exists():
        # path is latch:///a/b/ but b does not exist yet
        if not create_parents:
            raise ValueError(f"no such file or directory: {dest}")
        normalized = urljoins(normalized, src_path.name)

    if progress == Progress.none:
        num_bars = 0
        show_total_progress = False
    elif not src_path.is_dir():
        num_bars = 1
        show_total_progress = False
    else:
        num_bars = get_max_workers()
        show_total_progress = True

    with ProcessPoolExecutor(max_workers=get_max_workers()) as exec:
        with TransferStateManager() as man:
            parts_by_src: "PartsBySrcType" = man.dict()
            upload_info_by_src: "UploadInfoBySrcType" = man.dict()

            if src_path.is_dir():
                if dest_data.exists() and not src.endswith("/"):
                    normalized = urljoins(normalized, src_path.name)

                jobs: List[UploadJob] = []
                total_bytes = 0

                throttle: Throttle = man.Throttle()
                latency_q: "LatencyQueueType" = man.Queue()
                throttle_listener = exec.submit(throttler, throttle, latency_q)

                for dir_path, _, file_names in os.walk(src_path, followlinks=True):
                    for file_name in file_names:
                        rel_path = Path(dir_path) / file_name

                        try:
                            total_bytes += rel_path.stat().st_size
                        except FileNotFoundError:
                            print(f"WARNING: file {rel_path} not found, skipping...")
                            continue

                        parts_by_src[rel_path] = man.list()
                        jobs.append(
                            UploadJob(
                                rel_path,
                                urljoins(
                                    normalized,
                                    str(rel_path.relative_to(src_path)),
                                ),
                            )
                        )

                num_files = len(jobs)

                url_generation_bar: ProgressBars
                with closing(
                    man.ProgressBars(
                        0,
                        show_total_progress=(progress != Progress.none),
                    )
                ) as url_generation_bar:
                    url_generation_bar.set_total(num_files, "Generating URLs")

                    start_upload_futs: List[Future[Optional[StartUploadReturnType]]] = (
                        []
                    )

                    start = time.monotonic()
                    for job in jobs:
                        start_upload_futs.append(
                            exec.submit(
                                start_upload,
                                job.src,
                                job.dest,
                                url_generation_bar,
                                throttle,
                                latency_q,
                            )
                        )

                    for fut in as_completed(start_upload_futs):
                        res = fut.result()
                        if res is not None:
                            upload_info_by_src[res.src] = res

                    latency_q.put(None)
                    wait([throttle_listener])

                chunk_upload_bars: ProgressBars
                with closing(
                    man.ProgressBars(
                        min(num_bars, num_files),
                        show_total_progress=show_total_progress,
                        verbose=verbose,
                    )
                ) as chunk_upload_bars:
                    chunk_upload_bars.set_total(num_files, "Uploading Files")

                    # todo(ayush): async-ify
                    chunk_futs: List[Future[CompletedPart]] = []
                    for data in start_upload_futs:
                        res = data.result()

                        if res is None:
                            chunk_upload_bars.update_total_progress(1)
                            continue

                        pbar_index = chunk_upload_bars.get_free_task_bar_index()
                        chunk_upload_bars.set(
                            pbar_index,
                            res.src.stat().st_size,
                            res.src.name,
                        )
                        chunk_upload_bars.set_usage(str(res.src), res.part_count)

                        for part_index, url in enumerate(res.urls):
                            chunk_futs.append(
                                exec.submit(
                                    upload_file_chunk,
                                    src=res.src,
                                    url=url,
                                    part_index=part_index,
                                    part_size=res.part_size,
                                    progress_bars=chunk_upload_bars,
                                    pbar_index=pbar_index,
                                    parts_by_source=parts_by_src,
                                    upload_id=res.upload_id,
                                    dest=res.dest,
                                )
                            )

                    wait(chunk_futs)

                if progress != Progress.none:
                    print("\x1b[0GFinalizing uploads...")
            else:
                if dest_data.exists() and dest_is_dir:
                    normalized = urljoins(normalized, src_path.name)

                num_files = 1
                total_bytes = src_path.stat().st_size

                progress_bars: ProgressBars
                with closing(
                    man.ProgressBars(
                        num_bars,
                        show_total_progress=show_total_progress,
                        verbose=verbose,
                    )
                ) as progress_bars:
                    pbar_index = progress_bars.get_free_task_bar_index()

                    start = time.monotonic()
                    res = start_upload(src_path, normalized)

                    if res is not None:
                        progress_bars.set(
                            pbar_index, res.src.stat().st_size, res.src.name
                        )

                        chunk_futs: List[Future[CompletedPart]] = []
                        for part_index, url in enumerate(res.urls):
                            chunk_futs.append(
                                exec.submit(
                                    upload_file_chunk,
                                    src_path,
                                    url,
                                    part_index,
                                    res.part_size,
                                    progress_bars,
                                    pbar_index,
                                )
                            )

                        wait(chunk_futs)

                        end_upload(
                            normalized,
                            res.upload_id,
                            [fut.result() for fut in chunk_futs],
                        )

    end = time.monotonic()
    total_time = end - start

    return UploadResult(num_files, total_bytes, total_time)


@dataclass(frozen=True)
class StartUploadReturnType:
    upload_id: str
    urls: List[str]
    part_count: int
    part_size: int
    src: Path
    dest: str


MAX_RETRIES = 5


def start_upload(
    src: Path,
    dest: str,
    progress_bars: Optional[ProgressBars] = None,
    throttle: Optional[Throttle] = None,
    latency_q: Optional["LatencyQueueType"] = None,
) -> Optional[StartUploadReturnType]:
    if not src.exists():
        raise ValueError(f"could not find {src}: no such file or link")

    if src.is_symlink():
        src = src.resolve()

    content_type, _ = mimetypes.guess_type(src)
    if content_type is None:
        with open(src, "rb") as f:
            sample = f.read(units.KiB)

        try:
            sample.decode()
            content_type = "text/plain"
        except UnicodeDecodeError:
            content_type = "application/octet-stream"

    file_size = src.stat().st_size
    if file_size > latch_constants.maximum_upload_size:
        raise ValueError(
            f"file is {with_si_suffix(file_size)} which exceeds the maximum"
            " upload size (5TiB)",
        )

    part_count = min(
        latch_constants.maximum_upload_parts,
        math.ceil(file_size / latch_constants.file_chunk_size),
    )
    part_size = max(
        latch_constants.file_chunk_size,
        math.ceil(file_size / latch_constants.maximum_upload_parts),
    )

    if throttle is not None:
        time.sleep(throttle.get_delay())

    start = time.monotonic()
    res = tinyrequests.post(
        latch_config.api.data.start_upload,
        headers={"Authorization": get_auth_header()},
        json={
            "path": dest,
            "content_type": content_type,
            "part_count": part_count,
        },
        num_retries=MAX_RETRIES,
    )
    end = time.monotonic()

    if latency_q is not None:
        latency_q.put(end - start)

    json_data = res.json()

    if res.status_code != 200:
        raise RuntimeError(f"unable to start upload for {src}: {json_data['error']}")

    if progress_bars is not None:
        progress_bars.update_total_progress(1)

    if "version_id" in json_data["data"]:
        return  # file is empty, so no need to upload any content

    data: StartUploadData = json_data["data"]

    return StartUploadReturnType(
        **data, part_count=part_count, part_size=part_size, src=src, dest=dest
    )


@dataclass(frozen=True)
class CompletedPart:
    src: Path
    etag: str
    part_number: int


def upload_file_chunk(
    src: Path,
    url: str,
    part_index: int,
    part_size: int,
    progress_bars: Optional[ProgressBars] = None,
    pbar_index: Optional[int] = None,
    parts_by_source: Optional["PartsBySrcType"] = None,
    upload_id: Optional[str] = None,
    dest: Optional[str] = None,
) -> CompletedPart:
    time.sleep(0.1 * random.random())

    with open(src, "rb") as f:
        f.seek(part_size * part_index)
        data = f.read(part_size)

    res = tinyrequests.put(url, data=data)
    if res.status_code != 200:
        raise HTTPException(
            f"failed to upload part {part_index} of {src}: {res.status_code}"
        )

    ret = CompletedPart(
        src=src,
        etag=res.headers["ETag"],
        part_number=part_index + 1,
    )

    if parts_by_source is not None:
        parts_by_source[src].append(ret)

    if progress_bars is not None:
        progress_bars.update(pbar_index, len(data))
        pending_parts = progress_bars.dec_usage(str(src))

        if pending_parts == 0:
            progress_bars.return_task_bar(pbar_index)
            progress_bars.update_total_progress(1)
            progress_bars.write(f"Copied {src}")

            if (
                dest is not None
                and parts_by_source is not None
                and upload_id is not None
            ):
                end_upload(
                    dest=dest,
                    upload_id=upload_id,
                    parts=list(parts_by_source[src]),
                )

    return ret


def end_upload(
    dest: str,
    upload_id: str,
    parts: List[CompletedPart],
    progress_bars: Optional[ProgressBars] = None,
):
    res = tinyrequests.post(
        latch_config.api.data.end_upload,
        headers={"Authorization": get_auth_header()},
        json={
            "path": dest,
            "upload_id": upload_id,
            "parts": [
                {
                    "ETag": part.etag,
                    "PartNumber": part.part_number,
                }
                for part in parts
            ],
        },
    )

    if res.status_code != 200:
        err = res.json()["error"]
        if res.status_code == 400:
            raise ValueError(f"upload request invalid: {err}")
        if res.status_code == 401:
            raise RuntimeError(f"authorization failed: {err}")
        raise RuntimeError(
            f"end upload request failed with code {res.status_code}: {err}"
        )

    if progress_bars is not None:
        progress_bars.update_total_progress(1)


def throttler(t: Throttle, q: "LatencyQueueType"):
    ema = 0

    # todo(ayush): these params were tuned via naive grid search uploading a
    # test directory w/ ~19k files, ideally we should do something more rigorous
    alpha = 0.6
    beta = 1 / 60
    threshold = 15  # seconds

    while True:
        latency = q.get()
        if latency is None:
            return

        ema = (1 - alpha) * ema + alpha * latency

        if ema > threshold:
            t.set_delay(beta * ema)
        else:
            t.set_delay(0)



from typing import Dict, List, Union

from typing_extensions import TypeAlias

JsonArray: TypeAlias = List["JsonValue"]
"""JSON-compatible list"""

JsonObject: TypeAlias = Dict[str, "JsonValue"]
"""JSON-compatible dictionary"""

JsonValue: TypeAlias = Union[JsonObject, JsonArray, str, int, float, bool, None]
"""JSON-compatible value

Can be a dictionary, an array, or a primitive value
"""



import subprocess
import time
from concurrent.futures import ProcessPoolExecutor
from enum import Enum
from multiprocessing.managers import SyncManager
from pathlib import Path
from textwrap import dedent
from threading import Event
from typing import Dict, Optional

import click
from latch_sdk_config.latch import config

from latch_cli.constants import latch_constants
from latch_cli.menus import select_tui
from latch_cli.tinyrequests import post
from latch_cli.utils import (
    TemporarySSHCredentials,
    current_workspace,
    retrieve_or_login,
)

max_polls = 1800


class TaskSize(str, Enum):
    small_task = "small_task"
    medium_task = "medium_task"
    large_task = "large_task"
    small_gpu_task = "small_gpu_task"
    large_gpu_task = "large_gpu_task"


human_readable_task_sizes: Dict[str, TaskSize] = {
    "Small Task": TaskSize.small_task,
    "Medium Task": TaskSize.medium_task,
    "Large Task": TaskSize.large_task,
    "Small GPU Task": TaskSize.small_gpu_task,
    "Large GPU Task": TaskSize.large_gpu_task,
}


def _get_latest_image(pkg_root: Path) -> str:
    ws_id = current_workspace()
    if int(ws_id) < 10:
        ws_id = f"x{ws_id}"

    registry_name = f"{ws_id}_{pkg_root.name}"

    resp = post(
        config.api.workflow.get_latest,
        headers={"Authorization": f"Bearer {retrieve_or_login()}"},
        json={
            "registry_name": registry_name,
            "ws_account_id": current_workspace(),
        },
    )

    try:
        resp.raise_for_status()
        latest_version = resp.json()["version"]
    except:
        raise ValueError(
            "There was an issue getting your workflow's docker image. Please make sure"
            " you've registered your workflow at least once."
        )

    return f"{config.dkr_repo}/{ws_id}_{pkg_root.name}:{latest_version}"


def rsync(pkg_root: Path, ip: str, ssh_command: str, cancel_event: Event):
    while True:
        if cancel_event.is_set():
            return

        subprocess.run(
            [
                "rsync",
                f"--rsh={ssh_command}",
                "--compress",
                "--recursive",
                "--links",
                "--times",
                "--devices",
                "--specials",
                f"{pkg_root}/",
                f"root@{ip}:/root/",
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        # todo(ayush): use inotify or something instead of running on a fixed interval
        time.sleep(0.5)


def local_development(
    pkg_root: Path,
    *,
    skip_confirm_dialog: bool = False,
    size: Optional[TaskSize] = None,
    image: Optional[str] = None,
):
    """Starts a REPL that allows a user to interactively run tasks to help with
    debugging during workflow development.

    In this REPL, you can run tasks or scripts and make edits to them without
    having to reregister your workflow. You can also get a shell into a
    container with the same environment as the one that the workflow runs in, to
    help debug installation issues. See the full documentation for `Local
    Development` for more info.

    Like `get_executions`, this should only be called from the CLI for best
    results.

    Args:
        pkg_root: A path that points to a valid workflow directory (see the
            docs for `register`)

    """

    # ensure that rsync is installed
    try:
        subprocess.run(
            ["rsync", "--version"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )
    except FileNotFoundError:
        raise ValueError(dedent("""
                rsync is required for latch develop. Please install rsync and try again
                    linux: apt install rsync
                    mac: brew install rsync
                """))

    if image is None:
        image = _get_latest_image(pkg_root)

    if size is None:
        size = select_tui(
            "Please select an instance size",
            [
                {"display_name": k, "value": v}
                for k, v in human_readable_task_sizes.items()
            ],
        )
        if size is None:
            click.echo("Session cancelled.")
            return

    click.secho("Initiating local development session", fg="blue")
    click.echo(click.style("Selected image: ", fg="blue") + image)
    click.echo(click.style("Selected size: ", fg="blue") + size)

    if skip_confirm_dialog:
        click.echo("Proceeding without confirmation because of --yes")
    elif not click.confirm("Proceed?"):
        click.echo("Session cancelled.")
        return

    key_path = pkg_root / ".latch/ssh_key"
    jump_key_path = pkg_root / ".latch/jump_key"
    with TemporarySSHCredentials(key_path) as ssh:
        click.echo(
            "Starting local development session. This may take a few minutes for larger"
            " task sizes."
        )

        resp = post(
            "https://centromere.latch.bio/develop/start",
            headers={"Authorization": f"Latch-SDK-Token {retrieve_or_login()}"},
            json={
                "ImageName": image,
                "Workspace": current_workspace(),
                "SSHKey": ssh.public_key,
                "Size": size.value,
            },
        )

        json_data = resp.json()
        if resp.status_code != 200:
            raise ValueError(json_data["Error"])

        ip = json_data["IP"]
        jump_key = json_data["JumpKey"]

        jump_key_path.write_text(jump_key)
        jump_key_path.chmod(0o600)

        try:
            subprocess.run(
                ["ssh-add", str(jump_key_path)],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise ValueError("Unable to add jump host key to SSH Agent") from e

        try:
            poll_count = 0
            while poll_count < max_polls:
                resp = post(
                    "https://centromere.latch.bio/develop/ready",
                    headers={"Authorization": f"Latch-SDK-Token {retrieve_or_login()}"},
                )

                if resp.status_code != 200:
                    raise ValueError(resp.json()["Error"])

                if resp.json()["Ready"]:
                    break

                time.sleep(1)
                poll_count += 1

            if poll_count == max_polls:
                raise ValueError(
                    "Unable to provision instance due to server load - "
                    "please try again later."
                )

            ssh_command = [
                "ssh",
                "-o",
                "CheckHostIP=no",  # hack
                "-o",
                "StrictHostKeyChecking=no",
                "-o",
                "ServerAliveInterval=30",
                "-o",
                "ServerAliveCountMax=5",
                "-J",
                f"{latch_constants.jump_user}@{latch_constants.jump_host}",
                f"root@{ip}",
            ]

            with ProcessPoolExecutor() as exec:
                with SyncManager() as man:
                    stop_rsync = man.Event()
                    exec.submit(
                        rsync,
                        pkg_root,
                        ip,
                        " ".join(ssh_command[:-1]),
                        stop_rsync,
                    )

                    res = subprocess.run(ssh_command, stderr=subprocess.PIPE)
                    if "Too many authentication failures" in res.stderr.decode():
                        click.secho(
                            dedent("""
                            Too many authentication failures. Try resetting your ssh-agent with

                                $ ssh-add -D

                            and trying again."""),
                            fg="red",
                        )
                    stop_rsync.set()

        finally:
            resp = post(
                "https://centromere.latch.bio/develop/stop",
                headers={"Authorization": f"Latch-SDK-Token {retrieve_or_login()}"},
                json={"ImageName": image},
            )

            if resp.status_code != 200:
                raise ValueError(resp.json()["Error"])

            try:
                subprocess.run(
                    ["ssh-add", "-d", jump_key_path],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    check=True,
                )
            except subprocess.CalledProcessError as e:
                raise ValueError("Unable to remove jump host key from SSH Agent")

            jump_key_path.unlink(missing_ok=True)



import inspect
from typing import Any, Callable, Dict

from flytekit import LaunchPlan as _LaunchPlan


class LaunchPlan:
    """Construct named groups of default parameters for your workflows.

    Pass a workflow function and a dictionary of parameter names mapped to
    default python values and a set of "test data" will be populated on the
    console upon registration.

    Args:
        workflow: The workflow function to which the values apply.
        name: A semantic identifier to the parameter values (eg. 'Small Data')
        default_params: A mapping of paramter names to values

    ..
        from latch.resources.launch_plan import LaunchPlan

        LaunchPlan(
            assemble_and_sort,
            "foo",
            {"read1": LatchFile("latch:///foobar"), "read2": LatchFile("latch:///foobar")},
        )
    """

    def __init__(self, workflow: Callable, name: str, default_params: Dict[str, Any]):
        # This constructor is invoked twice when task code is executed.
        #   1. When the pyflyte-execute entrypoint is invoked to start task.
        #      `mod.__name__`  of caller is `wf`
        #   2. When the PythonAutoContainer loads our module to call our task.
        #      `mod.__name__`  of caller is `wf.__init__`
        # LaunchPlans are stored in a global array and redundant additions will
        # throw an exception, so we want to create on the first constructor
        # call only.

        frame = inspect.stack()[1]
        mod = inspect.getmodule(frame[0])
        if mod.__name__ == "wf":
            str_repr = f"wf.__init__.{workflow.__name__}.{name}"
            _LaunchPlan.create(str_repr, workflow, default_params)



import textwrap
from pathlib import Path
from typing import Optional

from ..services.register.utils import import_module_by_path


def load_snakemake_metadata(pkg_root: Path) -> Optional[Path]:
    new_meta = pkg_root / "latch_metadata" / "__init__.py"
    old_meta = pkg_root / "latch_metadata.py"

    if new_meta.exists():
        import_module_by_path(new_meta)

        return new_meta
    elif old_meta.exists():
        import_module_by_path(old_meta)

        return old_meta


# todo(maximsmol): use a stateful writer that keeps track of indent level
def reindent(x: str, level: int) -> str:
    if x[0] == "\n":
        x = x[1:]
    return textwrap.indent(textwrap.dedent(x), "    " * level)



import sys

from setuptools import find_packages, setup

cur_ver = sys.version_info[:2]
ver_str = ".".join(map(str, cur_ver))

if cur_ver < (3, 8) or cur_ver > (3, 11):
    raise RuntimeError(
        f"Python {ver_str} is unsupported. Please use a Python version between 3.8 and"
        " 3.11, inclusive."
    )

setup(
    name="latch",
    version="v2.39.7",
    author_email="kenny@latch.bio",
    description="The Latch SDK",
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.8,<3.12",
    entry_points={
        "console_scripts": [
            "latch=latch_cli.main:main",
        ]
    },
    install_requires=[
        "kubernetes>=24.2.0",
        "pyjwt>=0.2.0",
        "requests>=2.28.1",
        "click>=8.0",
        "docker>=5.0",
        "paramiko>=3.2.0",
        "scp>=0.14.0",
        "boto3>=1.26.0",
        "tqdm>=4.63.0",
        "lytekit==0.15.3",
        "lytekitplugins-pods==0.6.2",
        "typing-extensions==4.7.1",
        "apscheduler==3.9.1",
        "gql==3.4.0",
        "graphql-core==3.2.3",
        "requests-toolbelt==0.10.1",
        "latch-sdk-gql==0.0.6",
        "latch-sdk-config==0.0.4",
        "python-dateutil>=2.8",
        # for old latch develop, to be removed
        "aioconsole==0.6.1",
        "asyncssh==2.13.2",
        "websockets==11.0.3",
        "watchfiles==0.19.0",
    ],
    extras_require={
        "snakemake": ["snakemake>=7.18.0, <7.30.2", "pulp >=2.0, <2.8"],
        "pandas": ["pandas>=2.0.0"],
    },
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



"""Utility functions for services."""

import hashlib
import os
import re
import shutil
import stat
import subprocess
import urllib.parse
from dataclasses import asdict, is_dataclass
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from textwrap import dedent
from typing import List
from urllib.parse import urljoin

import click
import jwt
from latch_sdk_config.user import user_config

from latch_cli.click_utils import bold
from latch_cli.constants import latch_constants
from latch_cli.tinyrequests import get

# todo(ayush): need a better way to check if "latch" has been appended to urllib
if "latch" not in urllib.parse.uses_netloc:
    urllib.parse.uses_netloc.append("latch")
    urllib.parse.uses_relative.append("latch")


def urljoins(*args: str, dir: bool = False) -> str:
    """Construct a URL by appending paths

    Paths are always joined, with extra `/`s added if missing. Does not allow
    overriding basenames as opposed to normal `urljoin`. Whether the final
    path ends in a `/` is still significant and will be preserved in the output

    >>> urljoin("latch:///directory/", "another_directory")
    latch:///directory/another_directory
    >>> # No slash means "another_directory" is treated as a filename
    >>> urljoin(urljoin("latch:///directory/", "another_directory"), "file")
    latch:///directory/file
    >>> # Unintentionally overrode the filename
    >>> urljoins("latch:///directory/", "another_directory", "file")
    latch:///directory/another_directory/file
    >>> # Joined paths as expected

    Args:
        args: Paths to join
        dir: If true, ensure the output ends with a `/`
    """

    res = args[0]
    for x in args[1:]:
        if res[-1] != "/":
            res = f"{res}/"
        res = urljoin(res, x)

    if dir and res[-1] != "/":
        res = f"{res}/"

    return res


class AuthenticationError(RuntimeError): ...


def get_auth_header() -> str:
    sdk_token = user_config.token
    execution_token = os.environ.get("FLYTE_INTERNAL_EXECUTION_ID")

    if sdk_token is not None and sdk_token != "":
        header = f"Latch-SDK-Token {sdk_token}"
    elif execution_token is not None:
        header = f"Latch-Execution-Token {execution_token}"
    else:
        raise AuthenticationError("Unable to find authentication credentials.")

    return header


def retrieve_or_login() -> str:
    """Returns a valid JWT to access Latch, prompting a login flow if needed.

    Returns:
        A JWT
    """

    from latch_cli.services.login import login

    token = user_config.token
    if token == "":
        token = login()
    return token


def current_workspace() -> str:
    ws = user_config.workspace_id
    if ws == "":
        ws = account_id_from_token(retrieve_or_login())
        user_config.update_workspace(ws, "Personal Workspace")

    return ws


def sub_from_jwt(token: str) -> str:
    """Extract a user sub (UUID) from a JWT minted by auth0.

    Args:
        token: JWT

    Returns:
        The user sub contained within the JWT.
    """

    payload = jwt.decode(token, options={"verify_signature": False})
    try:
        sub = payload["sub"]
    except KeyError:
        raise ValueError(
            "Provided token lacks a user sub in the data payload"
            " and is not a valid token."
        )
    return sub


def account_id_from_token(token: str) -> str:
    """Exchanges a valid JWT for a Latch account ID.

    Latch account IDs are needed for any user-specific request, eg. register
    workflows or copy files to Latch.

    Args:
        token: JWT

    Returns:
        A Latch account ID (UUID).
    """

    decoded_jwt = jwt.decode(token, options={"verify_signature": False})
    try:
        return decoded_jwt.get("id")
    except KeyError as e:
        raise ValueError("Your Latch access token is malformed") from e


def _normalize_remote_path(remote_path: str):
    if remote_path.startswith("latch://"):
        remote_path = remote_path[len("latch://") :]
    if (
        not remote_path.startswith("/")
        and not remote_path.startswith("shared")
        and not remote_path.startswith("account")
    ):
        remote_path = f"/{remote_path}"

    return remote_path


def _si_number_strings(num):
    for unit in ["", "k", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1000:
            # `rstrip` remoes trailing zeroes
            return f"{num:3.1f}".rstrip("0").rstrip("."), unit
        num /= 1000
    return f"{num:.1f}", "Yi"


def with_si_suffix(num, suffix="B", styled=False):
    num, unit = _si_number_strings(num)

    if styled:
        import click

        return click.style(num, fg="bright_green", bold=True) + click.style(
            f"{unit}{suffix}", fg="green"
        )

    return f"{num}{unit}{suffix}"


def human_readable_time(t_seconds: float) -> str:
    s = t_seconds % 60
    m = (t_seconds // 60) % 60
    h = t_seconds // 60 // 60

    x: List[str] = []
    if h > 0:
        x.append(f"{int(h):d}h")
    if m > 0:
        x.append(f"{int(m):d}m")
    if s > 0:
        x.append(f"{s:.2f}s")

    return " ".join(x)


def hash_directory(dir_path: Path) -> str:
    # todo(maximsmol): store per-file hashes to show which files triggered a version change
    click.secho("Calculating workflow version based on file content hash", bold=True)
    click.secho("  Disable with --disable-auto-version/-d", italic=True, dim=True)

    m = hashlib.new("sha256")
    m.update(current_workspace().encode("utf-8"))

    ignore_file = dir_path / ".dockerignore"
    exclude: List[str] = ["/.latch", ".git"]
    try:
        with ignore_file.open("r") as f:
            click.secho("  Using .dockerignore", italic=True)

            for l in f:
                l = l.strip()

                if l == "":
                    continue
                if l[0] == "#":
                    continue

                exclude.append(l)
    except FileNotFoundError:
        ...

    from docker.utils import exclude_paths

    paths = list(exclude_paths(dir_path, exclude))
    paths.sort()

    for item in paths:
        p = Path(dir_path / item)

        p_stat = p.stat()
        if stat.S_ISDIR(p_stat.st_mode):
            m.update(p.name.encode("utf-8"))
            continue

        m.update(p.name.encode("utf-8"))

        file_size = p_stat.st_size
        if not stat.S_ISREG(p_stat.st_mode):
            click.secho(
                f"{p.relative_to(dir_path.resolve())} is not a regular file."
                " Ignoring contents",
                fg="yellow",
                bold=True,
            )
            continue

        if file_size > latch_constants.file_max_size:
            click.secho(
                f"{p.relative_to(dir_path.resolve())} is too large"
                f" ({with_si_suffix(file_size)}) to checksum. Ignoring contents",
                fg="yellow",
                bold=True,
            )
            continue

        m.update(p.read_bytes())

    return m.hexdigest()


def generate_temporary_ssh_credentials(
    ssh_key_path: Path, *, add_to_agent: bool = True
) -> str:
    # check if there is already a valid key at that path, and if so, use that
    # otherwise, if its not valid, remove it
    if ssh_key_path.exists():
        try:
            # check if file is valid + print out a fingerprint for the key
            cmd = ["ssh-keygen", "-l", "-f", ssh_key_path]
            valid_private_key = subprocess.run(cmd, check=True, capture_output=True)
            cmd = ["ssh-keygen", "-l", "-f", ssh_key_path.with_suffix(".pub")]
            valid_public_key = subprocess.run(cmd, check=True, capture_output=True)

            if valid_private_key.stdout != valid_public_key.stdout:
                raise

            # if both files are valid and their fingerprints match, use them instead of generating a new pair
            click.secho(
                f"Found existing key pair at {ssh_key_path}.", dim=True, italic=True
            )
        except:
            click.secho(
                f"Found malformed key-pair at {ssh_key_path}. Overwriting.",
                dim=True,
                italic=True,
            )

            ssh_key_path.unlink(missing_ok=True)
            ssh_key_path.with_suffix(".pub").unlink(missing_ok=True)

    # generate private key
    if not ssh_key_path.exists():
        ssh_key_path.parent.mkdir(parents=True, exist_ok=True)
        cmd = ["ssh-keygen", "-f", ssh_key_path, "-N", "", "-q"]

        try:
            subprocess.run(cmd, check=True)
        except subprocess.CalledProcessError as e:
            click.secho(
                dedent(f"""
                    There was a problem creating temporary SSH credentials. Please ensure
                    that `{bold("ssh-keygen")}` is installed and available in your PATH.
                """.strip()),
                fg="red",
            )

            raise click.exceptions.Exit(1) from e

        os.chmod(ssh_key_path, 0o700)

    if add_to_agent:
        # make key available to ssh-agent daemon
        cmd = ["ssh-add", ssh_key_path]

        try:
            subprocess.run(cmd, check=True, stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError as e:
            click.secho(
                dedent(f"""
                    There was an issue adding temporary SSH credentials to your SSH Agent.
                    Please activate your SSH Agent by running

                        {bold("$ eval `ssh-agent -s`")}

                    in your terminal.
                """.strip()),
                fg="red",
            )

            raise click.exceptions.Exit(1) from e

    # decode private key into public key
    cmd = ["ssh-keygen", "-y", "-f", ssh_key_path]
    try:
        out = subprocess.run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as e:
        click.secho(
            dedent(f"""
                There was a problem decoding your temporary credentials. Please ensure
                that `{bold("ssh-keygen")}` is installed and available in your PATH.
            """.strip()),
            fg="red",
        )

        raise click.exceptions.Exit(1) from e

    public_key = out.stdout.decode("utf-8").strip("\n")
    return public_key


def get_local_package_version() -> str:
    try:
        from importlib import metadata
    except ImportError:
        import importlib_metadata as metadata
    return metadata.version(latch_constants.pkg_name)


def get_latest_package_version_request() -> str:
    cache_location = user_config.root / "cached_version"
    resp = get(f"https://pypi.org/pypi/{latch_constants.pkg_name}/json")
    version = resp.json()["info"]["version"]
    with open(cache_location, "w") as f:
        f.write(f"{version} {datetime.now().isoformat()}")
    return version


def get_latest_package_version() -> str:
    version = None
    cache_location = user_config.root / "cached_version"
    try:
        with open(cache_location, "r") as f:
            version, timestamp = f.read().split(" ")
        if datetime.now() > datetime.fromisoformat(timestamp) + timedelta(days=1):
            version = get_latest_package_version_request()
    except FileNotFoundError:
        version = get_latest_package_version_request()

    return version


class TemporarySSHCredentials:
    def __init__(self, ssh_key_path: Path):
        self._ssh_key_path = ssh_key_path
        self._public_key = None

    def generate(self):
        if self._public_key is not None:
            return
        self._public_key = generate_temporary_ssh_credentials(self._ssh_key_path)

    def cleanup(self):
        if (
            self._ssh_key_path.exists()
            and self._ssh_key_path.with_suffix(".pub").exists()
        ):
            subprocess.run(
                ["ssh-add", "-d", self._ssh_key_path],
                check=True,
                capture_output=True,
            )
        self._ssh_key_path.unlink(missing_ok=True)
        self._ssh_key_path.with_suffix(".pub").unlink(missing_ok=True)

    @property
    def public_key(self):
        self.generate()
        return self._public_key

    @property
    def private_key(self):
        self.generate()
        with open(self._ssh_key_path, "r") as f:
            return f.read()

    def __enter__(self):
        self.generate()
        return self

    def __exit__(self, type, value, traceback):
        self.cleanup()


class WorkflowType(Enum):
    latchbiosdk = "latchbiosdk"
    snakemake = "snakemake"


def identifier_suffix_from_str(x: str) -> str:
    res = ""
    for c in x:
        res += c if f"_{c}".isidentifier() else "_"
    return res


def identifier_from_str(x: str) -> str:
    return (x[0] if x[0].isidentifier() else "_") + identifier_suffix_from_str(x[1:])


def get_parameter_json_value(v):
    if is_dataclass(v):
        return asdict(v)
    elif isinstance(v, Enum):
        return v.value
    elif isinstance(v, list):
        return [get_parameter_json_value(x) for x in v]
    elif isinstance(v, dict):
        return {k: get_parameter_json_value(x) for k, x in v.items()}
    else:
        return v


def check_exists_and_rename(old: Path, new: Path):
    if not new.exists():
        os.renames(old, new)
        return

    if old.is_file():
        if new.is_file():
            print(f"Warning: A file already exists at {new} and will be overwritten.")
            os.renames(old, new)
            return

        print(
            f"Warning: {old} is a file but {new} is not. Everything within {new} will"
            " be overwritten."
        )
        shutil.rmtree(new)
        os.renames(old, new)
        return

    if new.is_file():
        print(
            f"Warning: {old} is a directory but {new} is not. {new} will be"
            " overwritten."
        )
        shutil.rmtree(new)
        os.renames(old, new)
        return

    for sub in old.iterdir():
        check_exists_and_rename(sub, new / sub.name)


underscores = re.compile(r"_+")


def best_effort_display_name(x: str) -> str:
    return underscores.sub(" ", x).title().strip()



import os
import sys
import termios
import tty
import webbrowser
from pathlib import Path
from typing import List

from flytekit.core.workflow import PythonFunctionWorkflow
from google.protobuf.json_format import MessageToJson
from latch_sdk_config.latch import config

import latch_cli.menus as menus
from latch_cli.centromere.utils import _import_flyte_objects
from latch_cli.tinyrequests import post
from latch_cli.utils import current_workspace, retrieve_or_login


# TODO(ayush): make this import the `wf` directory and use the package root
# instead of the workflow name. also redo the frontend, also make it open the
# page
def preview(pkg_root: Path):
    """Generate a preview of the parameter interface for a workflow.

    This will allow a user to see how their parameter interface will look
    without having to first register their workflow.

    Args:
        pkg_root: A valid path pointing to the worklow code a user wishes to
            preview. The path can be absolute or relative.

    Example:
        >>> preview("wf.__init__.alphafold_wf")
    """

    try:
        modules = _import_flyte_objects([pkg_root.resolve()])
        wfs: dict[str, PythonFunctionWorkflow] = {}
        for module in modules:
            for flyte_obj in module.__dict__.values():
                if isinstance(flyte_obj, PythonFunctionWorkflow):
                    wfs[flyte_obj.name] = flyte_obj
        if len(wfs) == 0:
            raise ValueError(f"Unable to find a workflow definition in {pkg_root}")
    except ImportError as e:
        raise ValueError(
            f"Unable to find {e.name} - make sure that all necessary packages"
            " are installed and you have the correct function name."
        )

    wf = list(wfs.values())[0]
    if len(wfs) > 1:
        wf = wfs[
            _select_workflow_tui(
                title="Select which workflow to preview",
                options=list(wfs.keys()),
            )
        ]

    resp = post(
        url=config.api.workflow.preview,
        headers={"Authorization": f"Bearer {retrieve_or_login()}"},
        json={
            "workflow_ui_preview": MessageToJson(wf.interface.to_flyte_idl().inputs),
            "ws_account_id": current_workspace(),
        },
    )

    resp.raise_for_status()

    url = f"{config.console_url}/preview/parameters"
    webbrowser.open(url)


# TODO(ayush): abstract this logic in a unified interface that all tui commands use
def _select_workflow_tui(title: str, options: List[str], clear_terminal: bool = True):
    """
    Renders a terminal UI that allows users to select one of the options
    listed in `options`

    Args:
        title: The title of the selection window.
        options: A list of names for each of the options.
        clear_terminal: Whether or not to clear the entire terminal window
            before displaying - default False
    """

    if len(options) == 0:
        raise ValueError("No options given")

    def render(
        curr_selected: int,
        start_index: int = 0,
        max_per_page: int = 10,
        indent: str = "    ",
    ) -> int:
        if curr_selected < 0 or curr_selected >= len(options):
            curr_selected = 0

        menus._print(title)
        menus.line_down(2)

        num_lines_rendered = 4  # 4 "extra" lines for header + footer

        for i in range(start_index, start_index + max_per_page):
            if i >= len(options):
                break
            name = options[i]
            if i == curr_selected:
                color = "\x1b[38;5;40m"
                bold = "\x1b[1m"
                reset = "\x1b[0m"
                menus._print(f"{indent}{color}{bold}{name}{reset}\x1b[1E")
            else:
                menus._print(f"{indent}{name}\x1b[1E")
            num_lines_rendered += 1

        menus.line_down(1)

        control_str = "[ARROW-KEYS] Navigate\t[ENTER] Select\t[Q] Quit"
        menus._print(control_str)
        menus.line_up(num_lines_rendered - 1)

        menus._show()

        return num_lines_rendered

    old_settings = termios.tcgetattr(sys.stdin.fileno())
    tty.setraw(sys.stdin.fileno())

    curr_selected = 0
    start_index = 0
    _, term_height = os.get_terminal_size()
    menus.remove_cursor()

    if not clear_terminal:
        _, curs_height = menus.current_cursor_position()
        max_per_page = term_height - curs_height - 4
    else:
        menus.clear_screen()
        menus.move_cursor((0, 0))
        max_per_page = term_height - 4

    num_lines_rendered = render(
        curr_selected,
        start_index=start_index,
        max_per_page=max_per_page,
    )

    try:
        while True:
            b = menus.read_bytes(1)
            if b == b"\r":
                return options[curr_selected]
            elif b == b"\x1b":
                b = menus.read_bytes(2)
                if b == b"[A":  # Up Arrow
                    curr_selected = max(curr_selected - 1, 0)
                    if (
                        curr_selected - start_index < max_per_page // 2
                        and start_index > 0
                    ):
                        start_index -= 1
                elif b == b"[B":  # Down Arrow
                    curr_selected = min(curr_selected + 1, len(options) - 1)
                    if (
                        curr_selected - start_index > max_per_page // 2
                        and start_index < len(options) - max_per_page
                    ):
                        start_index += 1
                else:
                    continue
            menus.clear(num_lines_rendered)
            num_lines_rendered = render(
                curr_selected,
                start_index=start_index,
                max_per_page=max_per_page,
            )
    except KeyboardInterrupt: ...
    finally:
        menus.clear(num_lines_rendered)
        menus.reveal_cursor()
        menus._show()
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, old_settings)



"""Service to list test objects from a managed bucket."""

from typing import List

import boto3

from latch_cli.services.test_data.utils import _retrieve_creds

BUCKET = "latch-public"


def ls() -> List[str]:
    """Lists test data objects.

    Example:

        >>> ls()
            # Will return the full S3 paths of all of my objects.
    """

    session_token, access_key, secret_key, account_id = _retrieve_creds()

    s3_resource = boto3.resource(
        "s3",
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        aws_session_token=session_token,
    )

    objects = s3_resource.Bucket(BUCKET).objects.filter(
        Prefix=f"test-data/{account_id}"
    )
    return [x.key for x in objects]



"""Services or actions exposed through the SDK.

Services take one or more domain objects and/or values and perform some
transformation. These are the actions a user will interact with through the
CLI, eg. to "register" a workflow given login context.
"""



"""Service to remove test objects from a managed bucket."""

import boto3
import botocore

from latch_cli.services.test_data.utils import _retrieve_creds

BUCKET = "latch-public"


def remove(object_url: str):
    """Removes an object from a managed bucket.

    Args:
        object_url: The url of the object to remove.

    Example:

        >>> remove("s3://latch-public/1/foo.txt")
            # Removes the object at this path from your managed bucket prefix.
    """

    session_token, access_key, secret_key, account_id = _retrieve_creds()

    s3_resource = boto3.resource(
        "s3",
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        aws_session_token=session_token,
    )
    object_key = object_url[len(f"s3://{BUCKET}/") :]

    try:
        s3_resource.Object(BUCKET, object_key).load()
    except botocore.exceptions.ClientError:
        raise ValueError(f"{object_url} does not exist")

    s3_resource.Object(BUCKET, object_key).delete()



"""
Assemble and sort some COVID reads...
"""
from wf.assemble import assembly_task
from wf.sort import sort_bam_task

from latch.resources.launch_plan import LaunchPlan
from latch.resources.workflow import workflow
from latch.types.directory import LatchOutputDir
from latch.types.file import LatchFile
from latch.types.metadata import LatchAuthor, LatchMetadata, LatchParameter, LatchRule

"""The metadata included here will be injected into your interface."""
metadata = LatchMetadata(
    display_name="Assemble and Sort FastQ Files",
    documentation="your-docs.dev",
    author=LatchAuthor(
        name="Author",
        email="author@gmail.com",
        github="github.com/author",
    ),
    repository="https://github.com/your-repo",
    license="MIT",
    parameters={
        "read1": LatchParameter(
            display_name="Read 1",
            description="Paired-end read 1 file to be assembled.",
            batch_table_column=True,  # Show this parameter in batched mode.
            rules=[
                # validate the input file using regex
                LatchRule(
                    regex="(.fastq|.fastq.gz|.fq|.fq.gz)$",
                    message="Only fastq, fastq.gz, fq, fq.gz extensions are valid",
                )
            ],
        ),
        "read2": LatchParameter(
            display_name="Read 2",
            description="Paired-end read 2 file to be assembled.",
            batch_table_column=True,  # Show this parameter in batched mode.
            rules=[
                LatchRule(
                    regex="(.fastq|.fastq.gz|.fq|.fq.gz)$",
                    message="Only fastq, fastq.gz, fq, fq.gz extensions are valid",
                )
            ],
        ),
        "output_directory": LatchParameter(
            display_name="Output Directory",
            description="Where to place the result file.",
            batch_table_column=True,  # Show this parameter in batched mode.
        ),
    },
    tags=[],
)


@workflow(metadata)
def assemble_and_sort(
    read1: LatchFile, read2: LatchFile, output_directory: LatchOutputDir
) -> LatchFile:
    """Description...

    markdown header
    ----

    Write some documentation about your workflow in
    markdown here:

    > Regular markdown constructs work as expected.

    # Heading

    * content1
    * content2
    """
    sam = assembly_task(read1=read1, read2=read2, output_directory=output_directory)
    return sort_bam_task(sam=sam, output_directory=output_directory)


"""
Add test data with a LaunchPlan. Provide default values in a dictionary with
the parameter names as the keys. These default values will be available under
the 'Test Data' dropdown at console.latch.bio.
"""
LaunchPlan(
    assemble_and_sort,
    "Test Data",
    {
        "read1": LatchFile("s3://latch-public/init/r1.fastq"),
        "read2": LatchFile("s3://latch-public/init/r2.fastq"),
        "output_directory": LatchOutputDir("latch:///assemble_and_sort_outputs"),
    },
)



from copy import deepcopy
from typing import List, Optional, TypedDict

import click
import gql
import graphql
from latch_sdk_config.user import user_config
from latch_sdk_gql.execute import execute

from latch_cli.click_utils import color
from latch_cli.menus import select_tui
from latch_cli.utils import current_workspace


# todo(ayush): put this into latch_sdk_gql
def with_fragments(doc: graphql.DocumentNode, *fragments: graphql.DocumentNode):
    res = deepcopy(doc)

    defs = [*res.definitions]
    for fragment in fragments:
        defs.extend(fragment.definitions)

    res.definitions = tuple(defs)

    return res


egn_fragment = gql.gql("""
    fragment EGNFrag on ExecutionGraphNode {
        id
        status
        finalWorkflowGraphNode {
            nodeName
            taskInfo {
                displayName
            }
        }
        taskExecutionInfos {
            nodes {
                retry
            }
        }
        containerInfo(
            orderBy: INDEX_ASC
            filter: { status: { equalTo: RUNNING } }
        ) {
            nodes {
                containerName
                index
            }
        }
    }
""")


ei_fragment = gql.gql("""
    fragment EIFrag on ExecutionInfo {
        id
        status
        displayName
        workflow {
            displayName
        }
        executionGraphNodesByExecutionId(condition: { status: RUNNING }) {
            nodes {
                id
                ...EGNFrag
            }
        }
    }
""")


fragments = (egn_fragment, ei_fragment)


class ContainerNode(TypedDict):
    containerName: str
    index: str


class ContainerInfo(TypedDict):
    nodes: List[ContainerNode]


class TaskInfo(TypedDict):
    displayName: str


class FinalWGN(TypedDict):
    nodeName: str
    taskInfo: TaskInfo


class TaskExecutionNode(TypedDict):
    retry: str


class TaskExecutionInfos(TypedDict):
    nodes: List[TaskExecutionNode]


class EGNNode(TypedDict):
    id: str
    finalWorkflowGraphNode: FinalWGN
    taskExecutionInfos: TaskExecutionInfos
    containerInfo: ContainerInfo
    status: str


class EGNById(TypedDict):
    nodes: List[EGNNode]


class Workflow(TypedDict):
    displayName: str


class ExecutionInfoNode(TypedDict):
    id: str
    status: str
    displayName: str
    workflow: Workflow
    executionGraphNodesByExecutionId: EGNById


class RunningExecutions(TypedDict):
    nodes: List[ExecutionInfoNode]


def get_execution_info(execution_id: Optional[str]) -> ExecutionInfoNode:
    workspace_str: str = user_config.workspace_name or user_config.workspace_id

    if execution_id is not None:
        info: Optional[ExecutionInfoNode] = execute(
            with_fragments(
                gql.gql("""
                    query ExecutionInfo($executionId: BigInt!) {
                        executionInfo(id: $executionId) {
                            id
                            ...EIFrag
                        }
                    }
                """),
                *fragments,
            ),
            {"executionId": execution_id},
        )["executionInfo"]

        if info is None:
            click.secho(
                f"Could not find an execution with ID {color(execution_id)}.", fg="red"
            )
            raise click.exceptions.Exit(1)

        if info["status"] != "RUNNING":
            click.secho(
                f"The selected execution ({color(info['displayName'])}) is no longer"
                " running.",
                fg="red",
            )
            raise click.exceptions.Exit(1)

        return info

    res: RunningExecutions = execute(
        with_fragments(
            gql.gql("""
                query RunningExecutions($createdBy: BigInt!) {
                    runningExecutions(argWorkspaceId: $createdBy) {
                        nodes {
                            id
                            ...EIFrag
                        }
                    }
                }
            """),
            *fragments,
        ),
        {"createdBy": current_workspace()},
    )["runningExecutions"]

    if len(res["nodes"]) == 0:
        click.secho("You have no executions currently running.", dim=True)
        raise click.exceptions.Exit(0)

    if len(res["nodes"]) == 1:
        execution = res["nodes"][0]
        click.secho(
            "Selecting execution"
            f" {color(execution['displayName'])} as it is"
            " the only"
            " one currently running in Workspace"
            f" {color(workspace_str)}.",
        )

        return execution

    selected_execution = select_tui(
        "You have multiple executions running in this workspace"
        f" ({color(workspace_str)}). Which"
        " execution would you like to inspect?",
        [
            {
                "display_name": f'{x["displayName"]} ({x["workflow"]["displayName"]})',
                "value": x,
            }
            for x in res["nodes"]
        ],
        clear_terminal=False,
    )
    if selected_execution is None:
        click.secho("No execution selected. Exiting.", dim=True)
        raise click.exceptions.Exit(0)

    return selected_execution


def get_egn_info(
    execution_info: Optional[ExecutionInfoNode], egn_id: Optional[str] = None
) -> EGNNode:
    if egn_id is not None:
        res: Optional[EGNNode] = execute(
            with_fragments(
                gql.gql("""
                    query EGNInfo($egnId: BigInt!) {
                        executionGraphNode(id: $egnId) {
                            id
                            ...EGNFrag
                        }
                    }
                """),
                egn_fragment,
            ),
            {"egnId": egn_id},
        )["executionGraphNode"]

        if res is None:
            click.secho(f"Could not find a task with ID {color(egn_id)}.", fg="red")
            raise click.exceptions.Exit(1)

        if res["status"] != "RUNNING":
            click.secho(
                "The selected task"
                f" ({color(res['finalWorkflowGraphNode']['taskInfo']['displayName'])})"
                " is no longer running.",
                fg="red",
            )
            raise click.exceptions.Exit(1)

        click.secho(
            "Selecting task"
            f" {color(res['finalWorkflowGraphNode']['taskInfo']['displayName'])}.",
        )

        return res

    if execution_info is None:
        click.secho(
            "Aborting as neither an execution or task id were provided.", fg="red"
        )
        raise click.exceptions.Exit(1)

    egn_nodes = execution_info["executionGraphNodesByExecutionId"]["nodes"]

    if len(egn_nodes) == 0:
        click.secho(
            "No running tasks found for execution"
            f" {color(execution_info['displayName'])}.",
            fg="red",
        )
        raise click.exceptions.Exit(1)

    if len(egn_nodes) == 1:
        node = egn_nodes[0]
        click.secho(
            "Selecting task"
            f" {color(node['finalWorkflowGraphNode']['taskInfo']['displayName'])} as"
            " it is the only one running in Execution"
            f" {color(execution_info['displayName'])}.",
        )

        return egn_nodes[0]

    selected_egn_node = select_tui(
        f"The execution you selected ({color(execution_info['displayName'])}) has"
        " multiple running tasks. Which would you like to inspect?",
        [
            {
                "display_name": (
                    f"{x['finalWorkflowGraphNode']['taskInfo']['displayName']}"
                ),
                "value": x,
            }
            for x in egn_nodes
        ],
        clear_terminal=False,
    )

    if selected_egn_node is None:
        click.secho("No task selected.", dim=True)
        raise click.exceptions.Exit(0)

    click.secho(
        "Selecting task"
        f" {color(selected_egn_node['finalWorkflowGraphNode']['taskInfo']['displayName'])}.",
    )

    return selected_egn_node


def get_container_info(
    egn_info: EGNNode, container_index: Optional[int] = None
) -> Optional[ContainerNode]:
    container_infos = egn_info["containerInfo"]["nodes"]

    if container_index is not None:
        for container in container_infos:
            if int(container["index"]) != container_index:
                continue

            return container

        click.secho(
            f"The specified container index ({color(str(container_index))}) is either"
            " not present in this map task"
            f" ({color(egn_info['finalWorkflowGraphNode']['taskInfo']['displayName'])})"
            " or is no longer running."
        )
        raise click.exceptions.Exit(1)

    if len(container_infos) == 0:
        return None

    if len(container_infos) == 1:
        container = container_infos[0]
        click.echo(
            "Selecting container"
            f" {color(container['index'])} as it is the"
            " only running"
            " container in Map Task"
            f" {color(egn_info['finalWorkflowGraphNode']['taskInfo']['displayName'])}"
        )
        return container_infos[0]

    selected_container_info = select_tui(
        "The map task you selected"
        f" ({color(egn_info['finalWorkflowGraphNode']['taskInfo']['displayName'])}) has"
        " multiple running containers. Which one would you like to inspect?",
        [
            {"display_name": f'Container {x["index"]}', "value": x}
            for x in container_infos
        ],
        clear_terminal=False,
    )

    if selected_container_info is None:
        click.secho("No container selected.", dim=True)
        raise click.exceptions.Exit(0)

    return selected_container_info



from latch.verified.deseq2 import deseq2_wf
from latch.verified.mafft import mafft
from latch.verified.pathway import gene_ontology_pathway_analysis
from latch.verified.rnaseq import rnaseq
from latch.verified.trim_galore import trim_galore



import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
from pysam import VariantFile

quals = [record.qual for record in VariantFile(snakemake.input[0])]
plt.hist(quals)

plt.savefig(snakemake.output[0])



import contextlib
import re
import shutil
import sys
import tempfile
import time
import webbrowser
from pathlib import Path
from typing import Iterable, List, Optional

import click
import gql
import latch_sdk_gql.execute as l_gql
from scp import SCPClient

from ...centromere.ctx import _CentromereCtx
from ...centromere.utils import MaybeRemoteDir
from ...utils import WorkflowType, current_workspace
from ..workspace import _get_workspaces
from .constants import ANSI_REGEX, MAX_LINES
from .utils import (
    DockerBuildLogItem,
    build_image,
    register_serialized_pkg,
    serialize_pkg_in_container,
    upload_image,
)


def _delete_lines(num: int):
    """Deletes the previous len(lines) lines, assuming cursor is on a
    new line just below the first line to be deleted"""
    for i in range(num):
        click.echo("\x1b[1F\x1b[0G\x1b[2K", nl=False)


def _print_window(cur_lines: List[str], line: str):
    """Prints the lines curr_lines[1:] and line, overwriting curr_lines
    in the process"""
    if line == "":
        return cur_lines
    elif len(cur_lines) >= MAX_LINES:
        line = ANSI_REGEX.sub("", line)
        new_lines = cur_lines[len(cur_lines) - MAX_LINES + 1 :]
        new_lines.append(line)
        _delete_lines(len(cur_lines))
        for s in new_lines:
            click.echo("\x1b[38;5;245m" + s + "\x1b[0m")
        return new_lines
    else:
        click.echo("\x1b[38;5;245m" + line + "\x1b[0m")
        cur_lines.append(line)
        return cur_lines


docker_build_step_pat = re.compile("^Step [0-9]+/[0-9]+ :")


def print_and_write_build_logs(
    build_logs: Iterable[DockerBuildLogItem],
    image: str,
    pkg_root: Path,
    *,
    progress_plain: bool = False,
):
    click.secho(f"Building Docker image", bold=True)

    logs_path = pkg_root / ".latch" / ".logs" / image
    logs_path.mkdir(parents=True, exist_ok=True)

    click.echo(f"  Writing log to {click.style(logs_path, italic=True)}\n")

    with (logs_path / "docker-build-logs.txt").open("w") as save_file:
        cur_lines: list[str] = []

        for x in build_logs:
            # for dockerfile parse errors
            message = x.get("message")
            if message is not None:
                save_file.write(f"{message}\n")
                raise ValueError(message)

            lines = x.get("stream")
            error = x.get("error")
            if error is not None:
                save_file.write(f"{error}\n")
                click.secho(f"Error when building image:\n{error}", fg="red", bold=True)
                sys.exit(1)

            if lines is not None:
                save_file.write(f"{lines}\n")

                if not progress_plain:
                    for line in lines.split("\n"):
                        curr_terminal_width = shutil.get_terminal_size()[0]

                        if len(line) > curr_terminal_width:
                            line = line[: curr_terminal_width - 3] + "..."

                        if docker_build_step_pat.match(line):
                            _delete_lines(len(cur_lines))
                            cur_lines = []
                            click.secho(line, fg="blue")
                        else:
                            cur_lines = _print_window(cur_lines, line)
                else:
                    click.echo(lines, nl=False)

        if not progress_plain:
            _delete_lines(len(cur_lines))


def print_upload_logs(upload_image_logs, image):
    click.secho(f"Uploading Docker image", bold=True)
    prog_map = {}

    def _pp_prog_map(prog_map, prev_lines):
        if prev_lines > 0:
            click.echo("\x1b[2K\x1b[1E" * prev_lines + f"\x1b[{prev_lines}F", nl=False)
        prog_chunk = ""
        i = 0
        for id, prog in prog_map.items():
            if prog is None:
                continue
            prog_chunk += f"{id} ~ {prog}\n"
            i += 1
        if prog_chunk == "":
            return 0
        click.echo(prog_chunk + f"\x1b[{i}A", nl=False)
        return i

    prev_lines = 0

    for x in upload_image_logs:
        if (
            x.get("error") is not None
            and "denied: Your authorization token has expired." in x["error"]
        ):
            click.secho(
                f"\nDocker authorization token for {image} is expired.",
                fg="red",
                bold=True,
            )
            sys.exit(1)

        prog_map[x.get("id")] = x.get("progress")
        prev_lines = _pp_prog_map(prog_map, prev_lines)


def _print_reg_resp(resp, image):
    click.secho(f"Registering workflow {image}", bold=True)
    version = image.split(":")[1]

    if not resp.get("success"):
        error_str = f"Failed:\n\n"
        if resp.get("stderr") is not None:
            for line in resp.get("stderr").split("\n"):
                if not line:
                    continue

                if line.startswith('{"json"'):
                    continue

                error_str += line + "\n"

        if "task with different structure already exists" in error_str:
            error_str = (
                f"Version {version} already exists. Make sure that you've saved any"
                " changes you made."
            )

        click.secho(f"\n{error_str}", fg="red", bold=True)
        sys.exit(1)
    elif not "Successfully registered file" in resp["stdout"]:
        click.secho(
            f"\nVersion ({version}) already exists."
            " Make sure that you've saved any changes you made.",
            fg="red",
            bold=True,
        )
        sys.exit(1)

    click.echo(resp.get("stdout"))


def print_serialize_logs(serialize_logs, image):
    click.secho(f"\nSerializing workflow", bold=True)
    for x in serialize_logs:
        click.echo(x, nl=False)


def _build_and_serialize(
    ctx: _CentromereCtx,
    image_name: str,
    context_path: Path,
    tmp_dir: str,
    dockerfile: Optional[Path] = None,
    *,
    progress_plain: bool = False,
    cache_tasks: bool = False,
):
    assert ctx.pkg_root is not None

    jit_wf = None
    if ctx.workflow_type == WorkflowType.snakemake:
        assert ctx.snakefile is not None
        assert ctx.version is not None

        from ...snakemake.serialize import generate_jit_register_code
        from ...snakemake.workflow import build_jit_register_wrapper

        jit_wf = build_jit_register_wrapper(cache_tasks)
        generate_jit_register_code(
            jit_wf,
            ctx.pkg_root,
            ctx.snakefile,
            ctx.version,
            image_name,
            current_workspace(),
        )

    image_build_logs = build_image(ctx, image_name, context_path, dockerfile)
    print_and_write_build_logs(
        image_build_logs, image_name, ctx.pkg_root, progress_plain=progress_plain
    )

    if ctx.workflow_type == WorkflowType.snakemake:
        assert jit_wf is not None
        assert ctx.dkr_repo is not None

        from ...snakemake.serialize import serialize_jit_register_workflow

        serialize_jit_register_workflow(jit_wf, tmp_dir, image_name, ctx.dkr_repo)
    else:
        serialize_logs, container_id = serialize_pkg_in_container(
            ctx, image_name, tmp_dir
        )
        print_serialize_logs(serialize_logs, image_name)

        assert ctx.dkr_client is not None
        exit_status = ctx.dkr_client.wait(container_id)
        if exit_status["StatusCode"] != 0:
            click.secho("\nWorkflow failed to serialize", fg="red", bold=True)
            sys.exit(1)

        ctx.dkr_client.remove_container(container_id)

    click.echo()
    upload_image_logs = upload_image(ctx, image_name)
    print_upload_logs(upload_image_logs, image_name)


def _recursive_list(directory: Path) -> List[Path]:
    res: List[Path] = []

    stack: List[Path] = [directory]
    while len(stack) > 0:
        cur = stack.pop()
        for x in cur.iterdir():
            res.append(x)

            if x.is_dir():
                stack.append(x)

    return res


def register(
    pkg_root: str,
    *,
    disable_auto_version: bool = False,
    remote: bool = False,
    open: bool = False,
    skip_confirmation: bool = False,
    snakefile: Optional[Path] = None,
    progress_plain: bool = False,
    cache_tasks: bool = False,
    use_new_centromere: bool = False,
):
    """Registers a workflow, defined as python code, with Latch.

    Kicks off a three-legged OAuth2.0 flow outlined in `RFC6749`_.  Logic
    scaffolding this flow and detailed documentation can be found in the
    `latch.auth` package

    From a high-level, the user will be redirected to a browser and prompted to
    login. The SDK meanwhile spins up a callback server on a separate thread
    that will be hit when the browser login is successful with an access token.

    .. _RFC6749:
        https://datatracker.ietf.org/doc/html/rfc6749

    The major constituent steps are:

        - Constructing a Docker image
        - Serializing flyte objects within an instantiated container
        - Uploading the container with a latch-owned registry
        - Registering serialized objects + the container with latch.

    The Docker image is constructed by inferring relevant files + dependencies
    from the workflow package code itself. If a Dockerfile is provided
    explicitly, it will be used for image construction instead.

    The registration flow makes heavy use of `Flyte`_, and while the Latch SDK
    modifies many components to play nicely with Latch, eg. platform API,
    user-specific auth, the underlying concepts are nicely summarized in the
    `flytekit documentation`_.

    Args:
        pkg_root: A valid path pointing to the worklow code a user wishes to
            register. The path can be absolute or relative. The path is always
            a directory, with its structure exactly as constructed and
            described in the `cli.services.init` function.


    Example:
        >>> register("./example_workflow")

    .. _Flyte:
        https://docs.flyte.org
    .. _flytekit documentation:
        https://docs.flyte.org/en/latest/concepts/registration.html
    """

    if snakefile is not None:
        try:
            import snakemake
        except ImportError as e:
            click.secho("\n`snakemake` package is not installed.", fg="red", bold=True)
            sys.exit(1)

    with _CentromereCtx(
        Path(pkg_root),
        disable_auto_version=disable_auto_version,
        remote=remote,
        snakefile=snakefile,
        use_new_centromere=use_new_centromere,
    ) as ctx:
        assert ctx.workflow_name is not None, "Unable to determine workflow name"
        assert ctx.version is not None, "Unable to determine workflow version"

        # todo(maximsmol): we really want the workflow display name here
        click.echo(
            " ".join(
                [click.style("Workflow name:", fg="bright_blue"), ctx.workflow_name]
            )
        )
        click.echo(" ".join([click.style("Version:", fg="bright_blue"), ctx.version]))

        workspaces = _get_workspaces()
        ws_name = next(
            (
                x[1]
                for x in workspaces.items()
                if x[0] == current_workspace()
                or (current_workspace() == "" and x[1] == "Personal Workspace")
            ),
            "N/A",
        )
        click.echo(
            " ".join([
                click.style("Target workspace:", fg="bright_blue"),
                ws_name,
                f"({current_workspace()})",
            ])
        )
        click.echo(
            " ".join([
                click.style("Workflow root:", fg="bright_blue"),
                str(ctx.default_container.pkg_dir),
            ])
        )

        if use_new_centromere:
            click.secho("Using experimental registration server.", fg="yellow")

        if not skip_confirmation:
            if not click.confirm("Start registration?"):
                click.secho("Cancelled", bold=True)
                return
        else:
            click.secho("Skipping confirmation because of --yes", bold=True)

        click.secho("\nInitializing registration", bold=True)
        transport = None
        scp = None

        click.echo(
            " ".join([
                click.style("Docker Image:", fg="bright_blue"),
                ctx.default_container.image_name,
            ])
        )
        click.echo()

        if remote:
            click.secho("Connecting to remote server for docker build\n", bold=True)

            assert ctx.ssh_client is not None
            transport = ctx.ssh_client.get_transport()

            assert transport is not None
            scp = SCPClient(transport=transport, sanitize=lambda x: x)

        with contextlib.ExitStack() as stack:
            # We serialize locally with snakemake projects
            remote_dir_client = None
            if snakefile is None:
                remote_dir_client = ctx.ssh_client
            td: str = stack.enter_context(MaybeRemoteDir(remote_dir_client))
            _build_and_serialize(
                ctx,
                ctx.default_container.image_name,
                ctx.default_container.pkg_dir,
                td,
                dockerfile=ctx.default_container.dockerfile,
                progress_plain=progress_plain,
                cache_tasks=cache_tasks,
            )

            if remote and snakefile is None:
                local_td = Path(stack.enter_context(tempfile.TemporaryDirectory()))

                assert scp is not None
                scp.get(f"{td}/*", local_path=str(local_td), recursive=True)
            else:
                local_td = Path(td)

            protos = _recursive_list(local_td)

            for task_name, container in ctx.container_map.items():
                task_td = stack.enter_context(MaybeRemoteDir(ctx.ssh_client))
                try:
                    _build_and_serialize(
                        ctx,
                        container.image_name,
                        ctx.default_container.pkg_dir,
                        task_td,
                        dockerfile=container.dockerfile,
                        progress_plain=progress_plain,
                    )

                    if remote:
                        local_task_td = Path(
                            stack.enter_context(tempfile.TemporaryDirectory())
                        )

                        assert scp is not None
                        scp.get(
                            f"{task_td}/*", local_path=str(local_td), recursive=True
                        )

                        new_protos = _recursive_list(local_td)
                    else:
                        local_task_td = Path(task_td)

                    new_protos = _recursive_list(local_task_td)

                    try:
                        split_task_name = task_name.split(".")
                        task_name = ".".join(
                            split_task_name[split_task_name.index("wf") :]
                        )
                        for new_proto in new_protos:
                            if task_name in new_proto.name:
                                protos = [
                                    new_proto if new_proto.name == f.name else f
                                    for f in protos
                                ]
                    except ValueError as e:
                        raise ValueError(
                            f"Unable to match {task_name} to any of the protobuf files"
                            f" in {new_protos}"
                        ) from e
                except TypeError as e:
                    raise ValueError(
                        "The path to your provided dockerfile ",
                        f"{container.dockerfile} given to {task_name} is invalid.",
                    ) from e

            reg_resp = register_serialized_pkg(
                protos, ctx.token, ctx.version, current_workspace()
            )
            _print_reg_resp(reg_resp, ctx.default_container.image_name)

            click.secho("Successfully registered workflow.", fg="green", bold=True)

            wf_infos = []
            retries = 0

            wf_name = ctx.workflow_name
            while len(wf_infos) == 0:
                wf_infos = l_gql.execute(
                    gql.gql("""
                    query workflowQuery($name: String, $ownerId: BigInt, $version: String) {
                        workflowInfos(condition: { name: $name, ownerId: $ownerId, version: $version}) {
                            nodes {
                                id
                            }
                        }
                    }
                    """),
                    {
                        "name": wf_name,
                        "version": ctx.version,
                        "ownerId": current_workspace(),
                    },
                )["workflowInfos"]["nodes"]
                time.sleep(1)

                if retries >= 5:
                    click.secho(
                        "Failed to query workflow ID in 5 seconds.", fg="red", bold=True
                    )
                    click.secho(
                        "This could be due to high demand or a bug in the platform.",
                        fg="red",
                    )
                    click.secho(
                        "If the workflow is not visible in latch console, contact"
                        " support.",
                        fg="red",
                    )
                    break

                retries += 1

            if len(wf_infos) > 0:
                if len(wf_infos) > 1:
                    click.secho(
                        f"Workflow {ctx.workflow_name}:{ctx.version} is not unique."
                        " The link below might be wrong.",
                        fg="yellow",
                    )

                wf_id = wf_infos[0]["id"]
                url = f"https://console.latch.bio/workflows/{wf_id}"
                click.secho(url, fg="green")

                if open:
                    webbrowser.open(url)



import json
from typing import Tuple

from latch_sdk_config.latch import config

import latch_cli.tinyrequests as tinyrequests
from latch_cli.utils import current_workspace, retrieve_or_login


def _retrieve_creds() -> Tuple[str, str, str, str]:
    response = tinyrequests.post(
        config.api.data.test_data,
        headers={"Authorization": f"Bearer {retrieve_or_login()}"},
        json={
            "ws_account_id": current_workspace(),
        },
    )
    if response.status_code != 200:
        raise ValueError(
            "Unable to retrieve upload credentials. Server responded with"
            f" {response.json}."
        )

    try:
        output = response.json()
        session_token = output["tmp_session_token"]
        access_key = output["tmp_access_key"]
        secret_key = output["tmp_secret_key"]
    except (json.decoder.JSONDecodeError, AttributeError) as e:
        raise ValueError(
            "Malformed response from server attempting to retrieve upload credentials."
        ) from e

    account_id = current_workspace()

    return session_token, access_key, secret_key, account_id



from enum import Enum
from typing import Optional

from flytekit.core.launch_plan import reference_launch_plan

from ..types.directory import LatchDir
from ..types.file import LatchFile


class BaseQualityEncoding(Enum):
    phred33 = "--phred33"
    phred64 = "--phred64"


class AdapterSequence(Enum):
    auto = "auto"
    illumina = "--illumina"
    stranded_illumina = "--stranded_illumina"
    nextera = "--nextera"
    small_rna = "--small_rna"


@reference_launch_plan(
    project="1656",
    domain="development",
    name="wf.__init__.trim_galore",
    version="0.0.12-bcecec",
)
def trim_galore(
    input_forward: LatchFile,
    input_reverse: LatchFile,
    base_out: Optional[str],
    output_directory: Optional[LatchDir],
    fastqc_args: Optional[str],
    adapter: Optional[str],
    adapter2: Optional[str],
    consider_already_trimmed: Optional[int],
    max_length: Optional[int],
    max_n: Optional[float],
    clip_R1: Optional[int],
    clip_R2: Optional[int],
    three_prime_clip_R1: Optional[int],
    three_prime_clip_R2: Optional[int],
    hardtrim5: Optional[int],
    hardtrim3: Optional[int],
    quality: int = 20,
    base_quality_encoding: BaseQualityEncoding = BaseQualityEncoding.phred33,
    fastqc: bool = True,
    adapter_sequence: AdapterSequence = AdapterSequence.auto,
    stringency: int = 1,
    error_rate: float = 0.01,
    gzip_output_files: bool = False,
    length: int = 20,
    trim_n: bool = False,
    report_file: bool = True,
    polyA: bool = False,
    implicon: bool = False,
    retain_unpaired: bool = True,
    length_1: int = 35,
    length_2: int = 35,
) -> LatchDir: ...



import os
import re
from os import PathLike
from pathlib import Path
from typing import Optional, Type, Union
from urllib.parse import urlparse

import gql
from flytekit.core.annotation import FlyteAnnotation
from flytekit.core.context_manager import FlyteContext, FlyteContextManager
from flytekit.core.type_engine import TypeEngine, TypeTransformer
from flytekit.models.literals import Literal
from flytekit.types.file.file import FlyteFile, FlyteFilePathTransformer
from latch_sdk_gql.execute import execute
from typing_extensions import Annotated

from latch.types.utils import format_path, is_absolute_node_path, is_valid_url
from latch_cli.utils.path import normalize_path


class LatchFile(FlyteFile):
    """Represents a file object in the context of a task execution.

    The local path identifies the file object's location on local disk in
    the context of a task execution. `LatchFile` inherits implementation of
    `__fsopen__` from `FlyteFile`, so methods like `open` can retrieve a string
    representation of self.

    ..
        @task
        def task(file: LatchFile):

            with open(file, "r") as f:
                print(f.read())

            mypath = Path(file).resolve()


    The remote path identifies a remote location. The remote location, either a
    latch or s3 url, can be inspected from an object passed to the task to
    reveal its remote source.

    It can also to deposit the file to a latch path when the object is returned
    from a task.

    ..

        @task
        def task(file: LatchFile):

            path = file.remote_path # inspect remote location

            # Returning a different file to LatchData.
            return LatchFile("./foobar.txt", "latch:///foobar.txt")
    """

    def __init__(
        self,
        path: Union[str, PathLike],
        remote_path: Optional[Union[str, PathLike]] = None,
        **kwargs,
    ):
        if path is None:
            raise ValueError("Unable to instantiate LatchFile with None")

        # Cast PathLike objects so that LatchFile has consistent JSON
        # representation.
        parsed = urlparse(str(path))
        if parsed.scheme == "file":
            self.path = parsed.path
        elif parsed.scheme == "latch":
            self.path = normalize_path(str(path))
        else:
            self.path = str(path)

        self._path_generated = False

        if is_valid_url(self.path) and remote_path is None:
            self._remote_path = str(path)
        else:
            self._remote_path = None if remote_path is None else str(remote_path)

        if kwargs.get("downloader") is not None:
            super().__init__(self.path, kwargs["downloader"], self._remote_path)
        else:

            def downloader():
                ctx = FlyteContextManager.current_context()
                if (
                    ctx is not None
                    and hasattr(self, "_remote_path")
                    and self._remote_path is not None
                    # todo(kenny) is this necessary?
                    and ctx.inspect_objects_only is False
                ):
                    local_path_hint = self._remote_path
                    if is_absolute_node_path.match(self._remote_path) is not None:
                        data = execute(
                            gql.gql("""
                            query getName($argPath: String!) {
                                ldataResolvePathData(argPath: $argPath) {
                                    name
                                }
                            }
                            """),
                            {"argPath": self._remote_path},
                        )["ldataResolvePathData"]

                        if data is not None and data["name"] is not None:
                            local_path_hint = data["name"]

                    self._idempotent_set_path(local_path_hint)

                    return ctx.file_access.get_data(
                        self._remote_path,
                        self.path,
                        is_multipart=False,
                    )

            super().__init__(self.path, downloader, self._remote_path)

    def _idempotent_set_path(self, hint: Optional[str] = None):
        if self._path_generated:
            return

        ctx = FlyteContextManager.current_context()
        if ctx is None:
            return

        self.path = ctx.file_access.get_random_local_path(hint)
        self._path_generated = True

    def _create_imposters(self):
        self._idempotent_set_path()

        p = Path(self.path)
        p.parent.mkdir(exist_ok=True, parents=True)
        p.touch(exist_ok=True)

    @property
    def local_path(self) -> str:
        """File path local to the environment executing the task."""

        # This will manually download object to local disk in the case of a
        # user wishing to access the file without using `open`.
        self.__fspath__()

        return str(self.path)

    @property
    def remote_path(self) -> Optional[str]:
        """A url referencing in object in LatchData or s3."""
        return self._remote_path

    def __repr__(self):
        if self.remote_path is None:
            return f"LatchFile({repr(format_path(self.local_path))})"

        return (
            f"LatchFile({repr(self.path)},"
            f" remote_path={repr(format_path(self.remote_path))})"
        )

    def __str__(self):
        if self.remote_path is None:
            return "LatchFile()"
        return f"LatchFile({format_path(self.remote_path)})"


LatchOutputFile = Annotated[
    LatchFile,
    FlyteAnnotation(
        {"output": True},
    ),
]
"""A LatchFile tagged as the output of some workflow.

The Latch Console uses this metadata to avoid checking for existence of the
file at its remote path and displaying an error. This check is normally made to
avoid launching workflows with LatchFiles that point to objects that don't
exist.
"""


class LatchFilePathTransformer(FlyteFilePathTransformer):
    def __init__(self):
        TypeTransformer.__init__(self, name="LatchFilePath", t=LatchFile)

    def to_python_value(
        self,
        ctx: FlyteContext,
        lv: Literal,
        expected_python_type: Union[Type[LatchFile], PathLike],
    ) -> LatchFile:
        uri = lv.scalar.blob.uri
        if expected_python_type is PathLike:
            raise TypeError(
                "Casting from Pathlike to LatchFile is currently not supported."
            )

        if not issubclass(expected_python_type, LatchFile):
            raise TypeError(
                f"Neither os.PathLike nor LatchFile specified {expected_python_type}"
            )

        # This is a local file path, like /usr/local/my_file, don't mess with it. Certainly, downloading it doesn't
        # make any sense.
        if not ctx.file_access.is_remote(uri):
            return expected_python_type(uri)

        # For the remote case, return an FlyteFile object that can download
        local_path = ctx.file_access.get_random_local_path(uri)

        def _downloader():
            return ctx.file_access.get_data(uri, local_path, is_multipart=False)

        ret = LatchFile(local_path, uri, downloader=_downloader)
        ret._remote_source = uri
        return ret


TypeEngine.register(LatchFilePathTransformer())



from pathlib import Path
from typing import List, Optional

from latch.types.file import LatchFile
from latch.types.utils import is_valid_url


def file_glob(
    pattern: str, remote_directory: str, target_dir: Optional[Path] = None
) -> List[LatchFile]:
    """Constructs a list of LatchFiles from a glob pattern.

    Convenient utility for passing collections of files between tasks. See
    [nextflow's channels](https://www.nextflow.io/docs/latest/channel.html) or
    [snakemake's wildcards](https://snakemake.readthedocs.io/en/stable/snakefiles/rules.html#wildcards).
    for similar functionality in other orchestration tools.

    The remote location of each constructed LatchFile will be consructed by
    appending the file name returned by the pattern to the directory
    represented by the `remote_directory`.

    Args:
        pattern: A glob pattern to match a set of files, eg. '\*.py'. Will
            resolve paths with respect to the working directory of the caller.
        remote_directory: A valid latch URL pointing to a directory, eg.
            latch:///foo. This _must_ be a directory and not a file.
        target_dir: An optional Path object to define an alternate working
            directory for path resolution

    Returns:
        A list of instantiated LatchFile objects.

    Intended Use: ::

        @small_task
        def task():

            ...

            return file_glob("*.fastq.gz", "latch:///fastqc_outputs")

    """

    if not is_valid_url(remote_directory):
        return []

    if target_dir is None:
        wd = Path.cwd()
    else:
        wd = target_dir
    matched = sorted(wd.glob(pattern))

    return [LatchFile(str(file), remote_directory + file.name) for file in matched]
